# coding=utf-8
from __future__ import division

from ladybug.datacollection import HourlyDiscontinuousCollection, \
    HourlyContinuousCollection, MonthlyCollection, DailyCollection, \
    MonthlyPerHourCollection
from ladybug.datacollectionimmutable import HourlyDiscontinuousCollectionImmutable, \
    HourlyContinuousCollectionImmutable, MonthlyCollectionImmutable, \
    DailyCollectionImmutable, MonthlyPerHourCollectionImmutable
from ladybug.header import Header
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.dt import DateTime
from ladybug.datatype.temperature import Temperature
from ladybug.datatype.fraction import RelativeHumidity

import pytest
import sys
if (sys.version_info >= (3, 0)):
    xrange = range


def test_hourly():
    """Test the dicontinuous collections."""
    a_per = AnalysisPeriod(6, 21, 12, 6, 21, 13)
    dt1, dt2 = DateTime(6, 21, 12), DateTime(6, 21, 13)
    v1, v2 = 20, 25
    dc1 = HourlyDiscontinuousCollectionImmutable(
        Header(Temperature(), 'C', a_per), [v1, v2], [dt1, dt2])
    assert dc1.datetimes == (dt1, dt2)
    assert dc1.values == (v1, v2)
    assert not dc1.is_mutable
    with pytest.raises(AttributeError):
        dc1[0] = 18
    with pytest.raises(AttributeError):
        dc1.values = [18, 24]
    with pytest.raises(Exception):
        dc1.values.append(10)
    with pytest.raises(AttributeError):
        dc2 = dc1.convert_to_culled_timestep(1)

    dc2 = dc1.to_mutable()
    assert isinstance(dc2, HourlyDiscontinuousCollection)
    assert dc2.is_mutable
    dc2[0] = 18
    assert dc2[0] == 18
    dc2.values = [18, 24]
    assert dc2.values == (18, 24)
    with pytest.raises(Exception):
        dc2.values.append(10)  # make sure that we can still not append

    dc3 = dc2.to_immutable()
    assert isinstance(dc3, HourlyDiscontinuousCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc3.to_immutable()
    assert isinstance(dc4, HourlyDiscontinuousCollectionImmutable)
    assert not dc4.is_mutable


def test_init_continuous():
    """Test the init methods for continuous collections"""
    # Setup temperature data collection
    header = Header(Temperature(), 'C', AnalysisPeriod())
    values = list(xrange(8760))
    dc1 = HourlyContinuousCollectionImmutable(header, values)
    assert len(dc1.datetimes) == 8760
    assert list(dc1.values) == list(xrange(8760))
    assert not dc1.is_mutable
    with pytest.raises(AttributeError):
        dc1[0] = 18
    with pytest.raises(AttributeError):
        dc1.values = [24] * 8760
    with pytest.raises(Exception):
        dc1.values.append(10)
    with pytest.raises(AttributeError):
        dc2 = dc1.convert_to_culled_timestep(1)

    dc2 = dc1.to_mutable()
    assert isinstance(dc2, HourlyContinuousCollection)
    assert dc2.is_mutable
    dc2[0] = 18
    assert dc2[0] == 18
    dc2.values = [24] * 8760
    assert dc2.values == tuple([24] * 8760)
    with pytest.raises(Exception):
        dc2.values.append(10)  # make sure that we can still not append

    dc3 = dc2.to_immutable()
    assert isinstance(dc3, HourlyContinuousCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc3.to_immutable()
    assert isinstance(dc4, HourlyContinuousCollectionImmutable)
    assert not dc4.is_mutable


def test_daily():
    """Test the daily collections."""
    a_per = AnalysisPeriod(6, 21, 0, 6, 22, 23)
    v1, v2 = 20, 25
    dc1 = DailyCollectionImmutable(
        Header(Temperature(), 'C', a_per), [v1, v2], a_per.doys_int)
    assert dc1.datetimes == tuple(a_per.doys_int)
    assert dc1.values == (v1, v2)
    assert not dc1.is_mutable
    with pytest.raises(AttributeError):
        dc1[0] = 18
    with pytest.raises(AttributeError):
        dc1.values = [18, 24]
    with pytest.raises(Exception):
        dc1.values.append(10)

    dc2 = dc1.to_mutable()
    assert isinstance(dc2, DailyCollection)
    assert dc2.is_mutable
    dc2[0] = 18
    assert dc2[0] == 18
    dc2.values = [18, 24]
    assert dc2.values == (18, 24)
    with pytest.raises(Exception):
        dc2.values.append(10)  # make sure that we can still not append

    dc3 = dc2.to_immutable()
    assert isinstance(dc3, DailyCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc3.to_immutable()
    assert isinstance(dc4, DailyCollectionImmutable)
    assert not dc4.is_mutable


def test_monthly():
    """Test the monthly collections."""
    a_per = AnalysisPeriod(6, 1, 0, 7, 31, 23)
    v1, v2 = 20, 25
    dc1 = MonthlyCollectionImmutable(
        Header(Temperature(), 'C', a_per), [v1, v2], a_per.months_int)

    assert dc1.datetimes == tuple(a_per.months_int)
    assert dc1.values == (v1, v2)
    assert not dc1.is_mutable
    with pytest.raises(AttributeError):
        dc1[0] = 18
    with pytest.raises(AttributeError):
        dc1.values = [18, 24]
    with pytest.raises(Exception):
        dc1.values.append(10)

    dc2 = dc1.to_mutable()
    assert isinstance(dc2, MonthlyCollection)
    assert dc2.is_mutable
    dc2[0] = 18
    assert dc2[0] == 18
    dc2.values = [18, 24]
    assert dc2.values == (18, 24)
    with pytest.raises(Exception):
        dc2.values.append(10)  # make sure that we can still not append

    dc3 = dc2.to_immutable()
    assert isinstance(dc3, MonthlyCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc3.to_immutable()
    assert isinstance(dc4, MonthlyCollectionImmutable)
    assert not dc4.is_mutable


def test_monthly_per_hour():
    """Test the monthly per hour collections."""
    a_per = AnalysisPeriod(6, 1, 0, 7, 31, 23)
    vals = [20] * 24 + [25] * 24
    dc1 = MonthlyPerHourCollectionImmutable(
        Header(Temperature(), 'C', a_per), vals, a_per.months_per_hour)
    assert dc1.datetimes == tuple(a_per.months_per_hour)
    assert dc1.values == tuple(vals)
    assert not dc1.is_mutable
    with pytest.raises(AttributeError):
        dc1[0] = 18
    with pytest.raises(AttributeError):
        dc1.values = range(48)
    with pytest.raises(Exception):
        dc1.values.append(10)

    dc2 = dc1.to_mutable()
    assert isinstance(dc2, MonthlyPerHourCollection)
    assert dc2.is_mutable
    dc2[0] = 18
    assert dc2[0] == 18
    dc2.values = range(48)
    assert dc2.values == tuple(range(48))
    with pytest.raises(Exception):
        dc2.values.append(10)  # make sure that we can still not append

    dc3 = dc2.to_immutable()
    assert isinstance(dc3, MonthlyPerHourCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc3.to_immutable()
    assert isinstance(dc4, MonthlyPerHourCollectionImmutable)
    assert not dc4.is_mutable


def test_get_aligned_collection():
    """Test the method for getting an aligned discontinuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc1 = HourlyDiscontinuousCollection(header, values,
                                        header.analysis_period.datetimes)
    dc2 = dc1.get_aligned_collection(50, RelativeHumidity(), '%')
    assert dc2.header.data_type.name == 'Relative Humidity'
    assert dc2.header.unit == '%'
    assert isinstance(dc2, HourlyDiscontinuousCollection)
    assert dc2.is_mutable

    dc3 = dc1.get_aligned_collection(50, RelativeHumidity(), '%', mutable=False)
    assert dc3.header.data_type.name == 'Relative Humidity'
    assert dc3.header.unit == '%'
    assert isinstance(dc3, HourlyDiscontinuousCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc1.get_aligned_collection(50, RelativeHumidity(), '%', mutable=True)
    assert dc4.header.data_type.name == 'Relative Humidity'
    assert dc4.header.unit == '%'
    assert isinstance(dc4, HourlyDiscontinuousCollection)
    assert dc4.is_mutable


def test_get_aligned_collection_continuous():
    """Test the method for getting an aligned continuous collection."""
    header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    values = list(xrange(24))
    dc1 = HourlyContinuousCollection(header, values)
    dc2 = dc1.get_aligned_collection(50, RelativeHumidity(), '%')
    assert dc2.header.data_type.name == 'Relative Humidity'
    assert dc2.header.unit == '%'
    assert isinstance(dc2, HourlyContinuousCollection)
    assert dc2.is_mutable

    dc3 = dc1.get_aligned_collection(50, RelativeHumidity(), '%', mutable=False)
    assert dc3.header.data_type.name == 'Relative Humidity'
    assert dc3.header.unit == '%'
    assert isinstance(dc3, HourlyContinuousCollectionImmutable)
    assert not dc3.is_mutable

    dc4 = dc1.get_aligned_collection(50, RelativeHumidity(), '%', mutable=True)
    assert dc4.header.data_type.name == 'Relative Humidity'
    assert dc4.header.unit == '%'
    assert isinstance(dc4, HourlyContinuousCollection)
    assert dc4.is_mutable



# coding=utf-8
from ladybug.psychrometrics import humid_ratio_from_db_rh, enthalpy_from_db_hr, \
    wet_bulb_from_db_rh, dew_point_from_db_rh, rel_humid_from_db_hr, \
    rel_humid_from_db_enth, rel_humid_from_db_dpt, rel_humid_from_db_wb, \
    dew_point_from_db_hr, dew_point_from_db_enth, dew_point_from_db_wb, \
    db_temp_from_enth_hr, dew_point_from_db_rh_fast, wet_bulb_from_db_rh_fast

import pytest


def test_humid_ratio_from_db_rh():
    """Test the accuracy of the humid_ratio_from_db_rh function."""
    assert humid_ratio_from_db_rh(30, 0) == pytest.approx(0, rel=1e-3)
    assert humid_ratio_from_db_rh(30, 50) == pytest.approx(0.013314, rel=1e-3)
    assert humid_ratio_from_db_rh(30, 100) == pytest.approx(0.02721, rel=1e-3)
    assert humid_ratio_from_db_rh(20, 0) == pytest.approx(0, rel=1e-3)
    assert humid_ratio_from_db_rh(20, 50) == pytest.approx(0.00726, rel=1e-3)
    assert humid_ratio_from_db_rh(20, 100) == pytest.approx(0.014698, rel=1e-3)
    assert humid_ratio_from_db_rh(-20, 0) == pytest.approx(0, rel=1e-3)
    assert humid_ratio_from_db_rh(-20, 50) == pytest.approx(0.0003173, rel=1e-3)
    assert humid_ratio_from_db_rh(-20, 100) == pytest.approx(0.00063508, rel=1e-3)


def test_enthalpy_from_db_hr():
    """Test the accuracy of the enthalpy_from_db_hr function."""
    assert enthalpy_from_db_hr(30, 0) == pytest.approx(30.18, rel=1e-3)
    assert enthalpy_from_db_hr(30, 0.0133) == pytest.approx(64.18544, rel=1e-3)
    assert enthalpy_from_db_hr(30, 0.02721) == pytest.approx(99.750528, rel=1e-3)
    assert enthalpy_from_db_hr(20, 0) == pytest.approx(20.12, rel=1e-3)
    assert enthalpy_from_db_hr(20, 0.00726) == pytest.approx(38.547332, rel=1e-3)
    assert enthalpy_from_db_hr(20, 0.01469) == pytest.approx(57.406158, rel=1e-3)
    assert enthalpy_from_db_hr(-20, 0) == pytest.approx(0, rel=1e-3)
    assert enthalpy_from_db_hr(-20, 0.00031738) == pytest.approx(0, rel=1e-3)
    assert enthalpy_from_db_hr(-20, 0.000635) == pytest.approx(0, rel=1e-3)
    assert enthalpy_from_db_hr(-20, 0, -273.15) == pytest.approx(254.66889, rel=1e-3)
    assert enthalpy_from_db_hr(-20, 0.00031738, -273.15) == pytest.approx(255.6121, rel=1e-3)
    assert enthalpy_from_db_hr(-20, 0.000635, -273.15) == pytest.approx(256.556, rel=1e-3)


def test_dew_point_from_db_rh():
    """Test the accuracy of the dew_point_from_db_rh function."""
    assert dew_point_from_db_rh(30, 0) == pytest.approx(-273.15, rel=1e-3)
    assert dew_point_from_db_rh(30, 50) == pytest.approx(18.4466, rel=1e-3)
    assert dew_point_from_db_rh(30, 100) == pytest.approx(30, rel=1e-3)
    assert dew_point_from_db_rh(20, 0) == pytest.approx(-273.15, rel=1e-3)
    assert dew_point_from_db_rh(20, 50) == pytest.approx(9.27239, rel=1e-3)
    assert dew_point_from_db_rh(20, 100) == pytest.approx(20, rel=1e-3)
    assert dew_point_from_db_rh(-20, 0) == pytest.approx(-273.15, rel=1e-3)
    assert dew_point_from_db_rh(-20, 50) == pytest.approx(-27.0217, rel=1e-3)
    assert dew_point_from_db_rh(-20, 100) == pytest.approx(-20, rel=1e-3)

    assert dew_point_from_db_rh(180, 10) == pytest.approx(99.6844, rel=1e-3)
    assert dew_point_from_db_rh(180, 50) == pytest.approx(151.9373, rel=1e-3)
    assert dew_point_from_db_rh(180, 100) == pytest.approx(180, rel=1e-3)
    assert dew_point_from_db_rh(-80, 10) == pytest.approx(-93.065214, rel=1e-3)
    assert dew_point_from_db_rh(-80, 50) == pytest.approx(-84.125, rel=1e-3)
    assert dew_point_from_db_rh(-80, 100) == pytest.approx(-80, rel=1e-3)


def test_wet_bulb_from_db_rh():
    """Test the accuracy of the wet_bulb_from_db_rh function."""
    assert wet_bulb_from_db_rh(30, 0) == pytest.approx(10.49804, rel=1e-3)
    assert wet_bulb_from_db_rh(30, 50) == pytest.approx(22.011934, rel=1e-3)
    assert wet_bulb_from_db_rh(30, 100) == pytest.approx(30.0, rel=1e-3)
    assert wet_bulb_from_db_rh(20, 0) == pytest.approx(5.865, rel=1e-3)
    assert wet_bulb_from_db_rh(20, 50) == pytest.approx(13.7562, rel=1e-3)
    assert wet_bulb_from_db_rh(20, 100) == pytest.approx(20, rel=1e-3)
    assert wet_bulb_from_db_rh(-20, 0) == pytest.approx(-21.5142, rel=1e-3)
    assert wet_bulb_from_db_rh(-20, 50) == pytest.approx(-20.7405, rel=1e-3)
    assert wet_bulb_from_db_rh(-20, 100) == pytest.approx(-20, rel=1e-3)


def test_rel_humid_from_db_hr():
    """Test the accuracy of the rel_humid_from_db_hr function."""
    assert rel_humid_from_db_hr(30, 0) == pytest.approx(0, rel=1e-2)
    assert rel_humid_from_db_hr(30, 0.0133) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_hr(30, 0.02721) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_hr(20, 0) == pytest.approx(0, rel=1e-2)
    assert rel_humid_from_db_hr(20, 0.00726) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_hr(20, 0.01469) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_hr(-20, 0) == pytest.approx(0, rel=1e-2)
    assert rel_humid_from_db_hr(-20, 0.00031738) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_hr(-20, 0.000635) == pytest.approx(100, rel=1e-2)


def test_rel_humid_from_db_enth():
    """Test the accuracy of the rel_humid_from_db_enth function."""
    assert rel_humid_from_db_enth(30, 30.18) == pytest.approx(0, rel=1e-2)
    assert rel_humid_from_db_enth(30, 64.18544) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_enth(30, 99.750528) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_enth(20, 20.12) == pytest.approx(0, rel=1e-2)
    assert rel_humid_from_db_enth(20, 38.547332) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_enth(20, 57.406158) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_enth(-20, 0) > 100
    assert rel_humid_from_db_enth(-20, 255.6121, reference_temp=-273.15) == pytest.approx(50., rel=1e-2)
    assert rel_humid_from_db_enth(-20, 256.556, reference_temp=-273.15) == pytest.approx(100., rel=1e-2)


def test_rel_humid_from_db_dpt():
    """Test the accuracy of the rel_humid_from_db_dpt function."""
    assert rel_humid_from_db_dpt(30, 18.45805) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_dpt(30, 30) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_dpt(20, 9.270086) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_dpt(20, 20) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_dpt(-20, -27.0215503) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_dpt(-20, -20) == pytest.approx(100, rel=1e-2)


def test_rel_humid_from_db_wb():
    """Test the accuracy of the rel_humid_from_db_wb function."""
    assert rel_humid_from_db_wb(30, 10.4980) < 1
    assert rel_humid_from_db_wb(30, 22.01193) == pytest.approx(50, rel=1e-2)
    assert rel_humid_from_db_wb(30, 30.0) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_wb(20, 5.8649597) < 1
    assert rel_humid_from_db_wb(20, 13.756197) == pytest.approx(50, rel=1e-1)
    assert rel_humid_from_db_wb(20, 20) == pytest.approx(100, rel=1e-2)
    assert rel_humid_from_db_wb(-20, -21.51420288086) < 1
    assert 45 < rel_humid_from_db_wb(-20, -20.74057642) < 55
    assert rel_humid_from_db_wb(-20, -20) == pytest.approx(100, rel=1e-2)


def test_dew_point_from_db_hr():
    """Test the accuracy of the dew_point_from_db_hr function."""
    assert dew_point_from_db_hr(30, 0.015) == pytest.approx(20.330675, rel=1e-3)
    assert dew_point_from_db_hr(20, 0.01) == pytest.approx(14.0418, rel=1e-3)
    assert dew_point_from_db_hr(-20, 0.0003) == pytest.approx(-27.5661, rel=1e-3)


def test_dew_point_from_db_enth():
    """Test the accuracy of the dew_point_from_db_enth function."""
    assert dew_point_from_db_enth(30, 64.18544) == pytest.approx(18.43351, rel=1e-2)
    assert dew_point_from_db_enth(20, 38.547332) == pytest.approx(9.2678, rel=1e-2)
    assert dew_point_from_db_enth(-20, 0) == -20
    assert dew_point_from_db_enth(-20, 255.6121, reference_temp=-273.15) == pytest.approx(-27.01307, rel=1e-2)


def test_dew_point_from_db_wb():
    """Test the accuracy of the dew_point_from_db_wb function."""
    assert dew_point_from_db_wb(30, 22.144) == pytest.approx(18.593726, rel=1e-3)
    assert dew_point_from_db_wb(20, 13.88) == pytest.approx(9.35052249, rel=1e-3)
    assert dew_point_from_db_wb(-20, -20.84) == pytest.approx(-29.78065, rel=1e-3)


def test_db_temp_from_enth_hr():
    """Test the accuracy of the db_temp_from_enth_hr function."""
    assert db_temp_from_enth_hr(60, 0.015) == pytest.approx(21.74775, rel=1e-3)
    assert db_temp_from_enth_hr(60, 0.01) == pytest.approx(34.1499, rel=1e-3)
    assert db_temp_from_enth_hr(30, 0.005) == pytest.approx(17.23136, rel=1e-3)


def test_dew_point_from_db_rh_fast():
    """Test the accuracy of the dew_point_from_db_rh_fast function."""
    assert dew_point_from_db_rh_fast(30, 0) == pytest.approx(-273.15, rel=1e-3)
    assert dew_point_from_db_rh_fast(30, 50) == pytest.approx(18.45805, rel=1e-3)
    assert dew_point_from_db_rh_fast(30, 100) == pytest.approx(30, rel=1e-3)
    assert dew_point_from_db_rh_fast(20, 0) == pytest.approx(-273.15, rel=1e-3)
    assert dew_point_from_db_rh_fast(20, 50) == pytest.approx(9.270086, rel=1e-3)
    assert dew_point_from_db_rh_fast(20, 100) == pytest.approx(20, rel=1e-3)
    assert dew_point_from_db_rh_fast(-20, 0) == pytest.approx(-273.15, rel=1e-3)
    assert dew_point_from_db_rh_fast(-20, 50) == pytest.approx(-27.76753, rel=1e-3)
    assert dew_point_from_db_rh_fast(-20, 100) == pytest.approx(-20, rel=1e-3)


def test_wet_bulb_from_db_rh_fast():
    """Test the accuracy of the wet_bulb_from_db_rh_fast function."""
    assert wet_bulb_from_db_rh_fast(30, 0) == pytest.approx(10.871, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(30, 50) == pytest.approx(22.144, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(30, 100) == pytest.approx(29.0, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(20, 0) == pytest.approx(6.07, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(20, 50) == pytest.approx(13.88, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(20, 100) == pytest.approx(20, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(-20, 0) == pytest.approx(-21.69, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(-20, 50) == pytest.approx(-20.84, rel=1e-3)
    assert wet_bulb_from_db_rh_fast(-20, 100) == pytest.approx(-20, rel=1e-3)



# coding=utf-8
from __future__ import division

from ladybug.color import Color, Colorset, ColorRange

import pytest


def test_init_color():
    """Test the initialization of color objects."""
    color = Color(255, 0, 100)
    str(color)  # Test the color representation

    assert color.r == 255
    assert color.g == 0
    assert color.b == 100

    color.r = 100
    assert color.r == 100

    assert len(color) == 3
    assert color == Color(100, 0, 100)
    assert color != Color(100, 0, 0)
    assert color == color.duplicate()

    for c in color:
        assert isinstance(c, int)


def test_init_color_invalid():
    """Test the initialization of invalid color objects."""
    with pytest.raises(Exception):
        Color(256, 0, 100)
    with pytest.raises(Exception):
        Color(-1, 0, 100)
    with pytest.raises(Exception):
        Color(0, 256, 100)
    with pytest.raises(Exception):
        Color(0, -1, 100)
    with pytest.raises(Exception):
        Color(0, 0, 256)
    with pytest.raises(Exception):
        Color(0, 0, -1)


def test_from_dict():
    """Test the from dict method."""
    sample_dict = {'r': '255',
                   'g': '0',
                   'b': '100'}
    color = Color.from_dict(sample_dict)
    assert isinstance(color, Color)
    assert color.r == 255
    assert color.g == 0
    assert color.b == 100


def test_to_from_dict():
    """Test the to/from dict methods."""
    color = Color(255, 0, 100)
    color_dict = color.to_dict()
    new_color = Color.from_dict(color_dict)
    assert new_color.to_dict() == color_dict


def test_init_colorset():
    """Test the initialization of colorset objects."""
    colorset = Colorset()
    str(colorset)  # Test the color representation

    assert len(colorset) > 20
    assert len(colorset[0]) == 10
    for i in range(len(colorset)):
        assert len(colorset[i]) > 1

    for key in colorset._colors.keys():
        for col in colorset._colors[key]:
            assert len(col) == 3


def test_init_colorset_by_name():
    """Test the initialization of colorset objects by name."""
    assert len(Colorset.original()) > 1
    assert len(Colorset.nuanced()) > 1
    assert len(Colorset.multi_colored()) > 1
    assert len(Colorset.ecotect()) > 1
    assert len(Colorset.view_study()) > 1
    assert len(Colorset.shadow_study()) > 1
    assert len(Colorset.glare_study()) > 1
    assert len(Colorset.annual_comfort()) > 1
    assert len(Colorset.thermal_comfort()) > 1
    assert len(Colorset.thermal_comfort_utci()) > 1
    assert len(Colorset.heat_sensation()) > 1
    assert len(Colorset.cold_sensation()) > 1
    assert len(Colorset.benefit_harm()) > 1
    assert len(Colorset.harm()) > 1
    assert len(Colorset.benefit()) > 1
    assert len(Colorset.shade_benefit_harm()) > 1
    assert len(Colorset.shade_harm()) > 1
    assert len(Colorset.shade_benefit()) > 1
    assert len(Colorset.energy_balance()) > 1
    assert len(Colorset.energy_balance_storage()) > 1
    assert len(Colorset.therm()) > 1
    assert len(Colorset.cloud_cover()) > 1
    assert len(Colorset.black_to_white()) > 1
    assert len(Colorset.blue_green_red()) > 1
    assert len(Colorset.multicolored_2()) > 1
    assert len(Colorset.multicolored_3()) > 1


def test_init_color_range():
    """Test the initialization of color range objects."""
    color_range = ColorRange([Color(75, 107, 169), Color(245, 239, 103),
                              Color(234, 38, 0)])
    str(color_range)  # Test the color representation

    assert len(color_range) == 3
    assert isinstance(color_range.colors, tuple)
    assert isinstance(color_range.domain, tuple)
    assert color_range[0] == Color(75, 107, 169)


def test_color_range_discontinuous():
    """Test color range objects with discontinuous colors."""
    color_range = ColorRange(continuous_colors=False)
    color_range.domain = [100, 2000]
    color_range.colors = [Color(75, 107, 169), Color(245, 239, 103),
                          Color(234, 38, 0)]

    assert color_range.color(99) == Color(75, 107, 169)
    assert color_range.color(100) == Color(245, 239, 103)
    assert color_range.color(2000) == Color(245, 239, 103)
    assert color_range.color(2001) == Color(234, 38, 0)


def test_color_range_continuous():
    """Test color range objects with continuous colors."""
    color_range = ColorRange([Color(0, 100, 0), Color(100, 200, 100)], [0, 1000])

    assert color_range.color(-100) == Color(0, 100, 0)
    assert color_range.color(0) == Color(0, 100, 0)
    assert color_range.color(500) == Color(50, 150, 50)
    assert color_range.color(250) == Color(25, 125, 25)
    assert color_range.color(1000) == Color(100, 200, 100)
    assert color_range.color(1100) == Color(100, 200, 100)


def test_color_range_from_dict():
    """Test the from_dict method."""
    sample_dict = {'colors': [{'r': '0', 'g': '0', 'b': '0'},
                              {'r': '0', 'g': '255', 'b': '100'}]}
    color_range = ColorRange.from_dict(sample_dict)
    assert isinstance(color_range, ColorRange)
    assert color_range[0] == Color(0, 0, 0)
    assert color_range[1] == Color(0, 255, 100)


def test_color_range_to_from_dict():
    """Test the to/from dict methods."""
    color_range = ColorRange([Color(0, 100, 0), Color(100, 200, 100)], [0, 1000])
    color_range_dict = color_range.to_dict()
    new_color = ColorRange.from_dict(color_range_dict)
    assert new_color.to_dict() == color_range_dict



# coding=utf-8
from __future__ import division

from ladybug import datatype
from ladybug.datatype import base
from ladybug.datatype import angle, area, distance, energy, energyflux, \
    energyintensity, fraction, generic, illuminance, luminance, mass, massflowrate, \
    power, pressure, rvalue, speed, temperature, temperaturedelta, temperaturetime, \
    thermalcondition, specificenergy, uvalue, volume, volumeflowrate

import pytest
import math
PI = math.pi


def test_global_properties():
    """Test the global properties like TYPES, BASETYPES, and UNITS."""
    assert isinstance(datatype.TYPES, tuple)
    assert isinstance(datatype.BASETYPES, tuple)
    assert isinstance(datatype.UNITS, dict)
    assert isinstance(datatype.TYPESDICT, dict)


def test_from_dict():
    """Test the from dict method."""
    sample_dict = {'name': 'Temperature',
                   'data_type': 'Temperature',
                   'base_unit': 'C'}
    new_temp = base.DataTypeBase.from_dict(sample_dict)
    assert isinstance(new_temp, temperature.Temperature)


def test_to_from_dict():
    """Test the to/from dict methods."""
    my_temp = temperature.Temperature()
    temp_dict = my_temp.to_dict()
    new_temp = base.DataTypeBase.from_dict(temp_dict)
    new_temp = base.DataTypeBase.from_dict(temp_dict)
    assert new_temp.to_dict() == temp_dict


def test_dict_generic():
    """Test the from dict for a generic type."""
    sample_dict = {'name': 'Days Since Last Snowfall',
                   'data_type': 'GenericType',
                   'base_unit': 'days'}
    new_type = base.DataTypeBase.from_dict(sample_dict)
    assert isinstance(new_type, base.DataTypeBase)


def test_generic_type():
    """Test the creation of generic types."""
    test_type = generic.GenericType('Test Type', 'widgets')
    assert isinstance(test_type, base.DataTypeBase)
    assert test_type.is_unit_acceptable('widgets')


def test_temperature():
    """Test Temperature type."""
    temp_type = temperature.Temperature()
    for unit in temp_type.units:
        assert temp_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = temp_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = temp_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in temp_type.units:
            assert len(temp_type.to_unit([1], other_unit, unit)) == 1
    assert temp_type.to_unit([1], 'F', 'C')[0] == pytest.approx(33.8, rel=1e-1)
    assert temp_type.to_unit([1], 'K', 'C')[0] == pytest.approx(274.15, rel=1e-1)
    assert temp_type.to_unit([1], 'C', 'F')[0] == pytest.approx(-17.2222, rel=1e-1)
    assert temp_type.to_unit([1], 'C', 'K')[0] == pytest.approx(-272.15, rel=1e-1)


def test_temperaturedelta():
    """Test TemperatureDelta type."""
    temp_type = temperaturedelta.TemperatureDelta()
    for unit in temp_type.units:
        assert temp_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = temp_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = temp_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in temp_type.units:
            assert len(temp_type.to_unit([1], other_unit, unit)) == 1
    assert temp_type.to_unit([1], 'F', 'C')[0] == pytest.approx(1.8, rel=1e-1)
    assert temp_type.to_unit([1], 'K', 'C')[0] == pytest.approx(1, rel=1e-1)
    assert temp_type.to_unit([1], 'C', 'F')[0] == pytest.approx(0.5555, rel=1e-1)
    assert temp_type.to_unit([1], 'C', 'K')[0] == pytest.approx(1, rel=1e-1)


def test_temperaturetime():
    """Test TemperatureTime type."""
    temp_type = temperaturetime.TemperatureTime()
    for unit in temp_type.units:
        assert temp_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = temp_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = temp_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in temp_type.units:
            assert len(temp_type.to_unit([1], other_unit, unit)) == 1
    assert temp_type.to_unit([1], 'degF-days', 'degC-days')[0] == pytest.approx(1.8, rel=1e-1)
    assert temp_type.to_unit([1], 'degC-hours', 'degC-days')[0] == pytest.approx(24, rel=1e-3)
    assert temp_type.to_unit([1], 'degF-hours', 'degC-days')[0] == pytest.approx(43.2, rel=1e-1)
    assert temp_type.to_unit([1], 'degC-days', 'degF-days')[0] == pytest.approx(0.5555, rel=1e-1)
    assert temp_type.to_unit([1], 'degC-days', 'degC-hours')[0] == pytest.approx(1 / 24, rel=1e-1)
    assert temp_type.to_unit([1], 'degC-days', 'degF-hours')[0] == pytest.approx(0.023148, rel=1e-1)


def test_fraction():
    """Test Fraction type."""
    pct_type = fraction.Fraction()
    for unit in pct_type.units:
        assert pct_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = pct_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = pct_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in pct_type.units:
            assert len(pct_type.to_unit([1], other_unit, unit)) == 1
    assert pct_type.to_unit([1], 'fraction', '%')[0] == 0.01
    assert pct_type.to_unit([1], 'tenths', '%')[0] == 0.1
    assert pct_type.to_unit([1], 'thousandths', '%')[0] == 10
    assert pct_type.to_unit([1], '%', 'fraction')[0] == 100
    assert pct_type.to_unit([1], '%', 'tenths')[0] == 10
    assert pct_type.to_unit([1], '%', 'thousandths')[0] == 0.1


def test_distance():
    """Test Distance type."""
    dist_type = distance.Distance()
    for unit in dist_type.units:
        assert dist_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = dist_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = dist_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in dist_type.units:
            assert len(dist_type.to_unit([1], other_unit, unit)) == 1
    assert dist_type.to_unit([1], 'ft', 'm')[0] == pytest.approx(3.28084, rel=1e-2)
    assert dist_type.to_unit([1], 'mm', 'm')[0] == 1000
    assert dist_type.to_unit([1], 'in', 'm')[0] == pytest.approx(39.3701, rel=1e-1)
    assert dist_type.to_unit([1], 'km', 'm')[0] == 0.001
    assert dist_type.to_unit([1], 'mi', 'm')[0] == pytest.approx(1 / 1609.344, rel=1e-1)
    assert dist_type.to_unit([1], 'cm', 'm')[0] == 100
    assert dist_type.to_unit([1], 'm', 'ft')[0] == pytest.approx(1 / 3.28084, rel=1e-2)
    assert dist_type.to_unit([1], 'm', 'mm')[0] == 0.001
    assert dist_type.to_unit([1], 'm', 'in')[0] == pytest.approx(1 / 39.3701, rel=1e-1)
    assert dist_type.to_unit([1], 'm', 'km')[0] == 1000
    assert dist_type.to_unit([1], 'm', 'mi')[0] == pytest.approx(1609.344, rel=1e-1)
    assert dist_type.to_unit([1], 'm', 'cm')[0] == 0.01


def test_area():
    """Test Area type."""
    area_type = area.Area()
    for unit in area_type.units:
        assert area_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = area_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = area_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in area_type.units:
            assert len(area_type.to_unit([1], other_unit, unit)) == 1
    assert area_type.to_unit([1], 'ft2', 'm2')[0] == pytest.approx(10.7639, rel=1e-2)
    assert area_type.to_unit([1], 'mm2', 'm2')[0] == 1000000
    assert area_type.to_unit([1], 'in2', 'm2')[0] == pytest.approx(1550, rel=1e-1)
    assert area_type.to_unit([1], 'km2', 'm2')[0] == 0.000001
    assert area_type.to_unit([1], 'mi2', 'm2')[0] == pytest.approx(1 / 2590000, rel=1e-8)
    assert area_type.to_unit([1], 'cm2', 'm2')[0] == 10000
    assert area_type.to_unit([1], 'ha', 'm2')[0] == 0.0001
    assert area_type.to_unit([1], 'acre', 'm2')[0] == pytest.approx(1 / 4046.86, rel=1e-8)
    assert area_type.to_unit([1], 'm2', 'ft2')[0] == pytest.approx(1 / 10.7639, rel=1e-3)
    assert area_type.to_unit([1], 'm2', 'mm2')[0] == 0.000001
    assert area_type.to_unit([1], 'm2', 'in2')[0] == pytest.approx(1 / 1550, rel=1e-4)
    assert area_type.to_unit([1], 'm2', 'km2')[0] == 1000000
    assert area_type.to_unit([1], 'm2', 'mi2')[0] == pytest.approx(2590000, rel=1e-1)
    assert area_type.to_unit([1], 'm2', 'cm2')[0] == 0.0001
    assert area_type.to_unit([1], 'm2', 'ha')[0] == 10000
    assert area_type.to_unit([1], 'm2', 'acre')[0] == pytest.approx(4046.86, rel=1e-1)


def test_volume():
    """Test Volume type."""
    vol_type = volume.Volume()
    for unit in vol_type.units:
        assert vol_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = vol_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = vol_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in vol_type.units:
            assert len(vol_type.to_unit([1], other_unit, unit)) == 1
    assert vol_type.to_unit([1], 'ft3', 'm3')[0] == pytest.approx(35.3147, rel=1e-2)
    assert vol_type.to_unit([1], 'mm3', 'm3')[0] == 1e+9
    assert vol_type.to_unit([1], 'in3', 'm3')[0] == pytest.approx(61023.7, rel=1e-1)
    assert vol_type.to_unit([1], 'km3', 'm3')[0] == 0.000000001
    assert vol_type.to_unit([1], 'mi3', 'm3')[0] == pytest.approx(1 / 4.168e+9, rel=1e-8)
    assert vol_type.to_unit([1], 'L', 'm3')[0] == 1000
    assert vol_type.to_unit([1], 'mL', 'm3')[0] == 1000000
    assert vol_type.to_unit([1], 'gal', 'm3')[0] == pytest.approx(264.172, rel=1e-1)
    assert vol_type.to_unit([1], 'fl oz', 'm3')[0] == pytest.approx(33814, rel=1e-1)
    assert vol_type.to_unit([1], 'm3', 'ft3')[0] == pytest.approx(1 / 35.3147, rel=1e-3)
    assert vol_type.to_unit([1], 'm3', 'mm3')[0] == 0.000000001
    assert vol_type.to_unit([1], 'm3', 'in3')[0] == pytest.approx(1 / 61023.7, rel=1e-5)
    assert vol_type.to_unit([1], 'm3', 'km3')[0] == 1e+9
    assert vol_type.to_unit([1], 'm3', 'mi3')[0] == pytest.approx(4.168e+9, rel=1e-1)
    assert vol_type.to_unit([1], 'm3', 'L')[0] == 0.001
    assert vol_type.to_unit([1], 'm3', 'mL')[0] == 0.000001
    assert vol_type.to_unit([1], 'm3', 'gal')[0] == pytest.approx(1 / 264.172, rel=1e-4)
    assert vol_type.to_unit([1], 'm3', 'fl oz')[0] == pytest.approx(1 / 33814, rel=1e-5)


def test_pressure():
    """Test Pressure type."""
    press_type = pressure.Pressure()
    for unit in press_type.units:
        assert press_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = press_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = press_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in press_type.units:
            assert len(press_type.to_unit([1], other_unit, unit)) == 1
    assert press_type.to_unit([1], 'inHg', 'Pa')[0] == pytest.approx(0.0002953, rel=1e-5)
    assert press_type.to_unit([1], 'atm', 'Pa')[0] == pytest.approx(1 / 101325, rel=1e-7)
    assert press_type.to_unit([1], 'bar', 'Pa')[0] == 0.00001
    assert press_type.to_unit([1], 'Torr', 'Pa')[0] == pytest.approx(0.00750062, rel=1e-5)
    assert press_type.to_unit([1], 'psi', 'Pa')[0] == pytest.approx(0.000145038, rel=1e-5)
    assert press_type.to_unit([1], 'inH2O', 'Pa')[0] == pytest.approx(0.00401865, rel=1e-5)
    assert press_type.to_unit([1], 'Pa', 'inHg')[0] == pytest.approx(1 / 0.0002953, rel=1e-1)
    assert press_type.to_unit([1], 'Pa', 'atm')[0] == pytest.approx(101325, rel=1e-1)
    assert press_type.to_unit([1], 'Pa', 'bar')[0] == 100000
    assert press_type.to_unit([1], 'Pa', 'Torr')[0] == pytest.approx(1 / 0.00750062, rel=1e-1)
    assert press_type.to_unit([1], 'Pa', 'psi')[0] == pytest.approx(1 / 0.000145038, rel=1e-1)
    assert press_type.to_unit([1], 'Pa', 'inH2O')[0] == pytest.approx(1 / 0.00401865, rel=1e-1)


def test_energy():
    """Test Energy type."""
    energy_type = energy.Energy()
    for unit in energy_type.units:
        assert energy_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = energy_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = energy_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in energy_type.units:
            assert len(energy_type.to_unit([1], other_unit, unit)) == 1
    assert energy_type.to_unit([1], 'kBtu', 'kWh')[0] == pytest.approx(3.41214, rel=1e-2)
    assert energy_type.to_unit([1], 'Wh', 'kWh')[0] == 1000
    assert energy_type.to_unit([1], 'Btu', 'kWh')[0] == pytest.approx(3412.14, rel=1e-1)
    assert energy_type.to_unit([1], 'MMBtu', 'kWh')[0] == pytest.approx(0.00341214, rel=1e-5)
    assert energy_type.to_unit([1], 'J', 'kWh')[0] == 3600000
    assert energy_type.to_unit([1], 'kJ', 'kWh')[0] == 3600
    assert energy_type.to_unit([1], 'MJ', 'kWh')[0] == 3.6
    assert energy_type.to_unit([1], 'GJ', 'kWh')[0] == 0.0036
    assert energy_type.to_unit([1], 'therm', 'kWh')[0] == pytest.approx(0.0341214, rel=1e-5)
    assert energy_type.to_unit([1], 'cal', 'kWh')[0] == pytest.approx(860421, rel=1e-1)
    assert energy_type.to_unit([1], 'kcal', 'kWh')[0] == pytest.approx(860.421, rel=1e-2)
    assert energy_type.to_unit([1], 'kWh', 'kBtu')[0] == pytest.approx(1 / 3.41214, rel=1e-2)
    assert energy_type.to_unit([1], 'kWh', 'Wh')[0] == 0.001
    assert energy_type.to_unit([1], 'kWh', 'Btu')[0] == pytest.approx(1 / 3412.14, rel=1e-5)
    assert energy_type.to_unit([1], 'kWh', 'MMBtu')[0] == pytest.approx(1 / 0.00341214, rel=1e-1)
    assert energy_type.to_unit([1], 'kWh', 'J')[0] == 1 / 3600000
    assert energy_type.to_unit([1], 'kWh', 'kJ')[0] == 1 / 3600
    assert energy_type.to_unit([1], 'kWh', 'MJ')[0] == 1 / 3.6
    assert energy_type.to_unit([1], 'kWh', 'GJ')[0] == 1 / 0.0036
    assert energy_type.to_unit([1], 'kWh', 'therm')[0] == pytest.approx(1 / 0.0341214, rel=1e-1)
    assert energy_type.to_unit([1], 'kWh', 'cal')[0] == pytest.approx(1 / 860421, rel=1e-7)
    assert energy_type.to_unit([1], 'kWh', 'kcal')[0] == pytest.approx(1 / 860.421, rel=1e-4)


def test_energy_intensity():
    """Test Energy type."""
    energyi_type = energyintensity.EnergyIntensity()
    for unit in energyi_type.units:
        assert energyi_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = energyi_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = energyi_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in energyi_type.units:
            assert len(energyi_type.to_unit([1], other_unit, unit)) == 1
    assert energyi_type.to_unit([1], 'kBtu/ft2', 'kWh/m2')[0] == pytest.approx(0.316998, rel=1e-3)
    assert energyi_type.to_unit([1], 'Wh/m2', 'kWh/m2')[0] == 1000
    assert energyi_type.to_unit([1], 'Btu/ft2', 'kWh/m2')[0] == pytest.approx(316.998, rel=1e-1)
    assert energyi_type.to_unit([1], 'kWh/m2', 'kBtu/ft2')[0] == pytest.approx(1 / 0.316998, rel=1e-2)
    assert energyi_type.to_unit([1], 'kWh/m2', 'Wh/m2')[0] == 0.001
    assert energyi_type.to_unit([1], 'kWh/m2', 'Btu/ft2')[0] == pytest.approx(1 / 316.998, rel=1e-5)


def test_power():
    """Test Power type."""
    power_type = power.Power()
    for unit in power_type.units:
        assert power_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = power_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = power_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in power_type.units:
            assert len(power_type.to_unit([1], other_unit, unit)) == 1
    assert power_type.to_unit([1], 'Btu/h', 'W')[0] == pytest.approx(3.41214, rel=1e-3)
    assert power_type.to_unit([1], 'kW', 'W')[0] == 0.001
    assert power_type.to_unit([1], 'kBtu/h', 'W')[0] == pytest.approx(0.00341214, rel=1e-5)
    assert power_type.to_unit([1], 'TR', 'W')[0] == pytest.approx(1 / 3516.85, rel=1e-5)
    assert power_type.to_unit([1], 'hp', 'W')[0] == pytest.approx(1 / 745.7, rel=1e-5)
    assert power_type.to_unit([1], 'W', 'Btu/h')[0] == pytest.approx(1 / 3.41214, rel=1e-2)
    assert power_type.to_unit([1], 'W', 'kW')[0] == 1000
    assert power_type.to_unit([1], 'W', 'kBtu/h')[0] == pytest.approx(1 / 0.00341214, rel=1e-1)
    assert power_type.to_unit([1], 'W', 'TR')[0] == pytest.approx(3516.85, rel=1e-1)
    assert power_type.to_unit([1], 'W', 'hp')[0] == pytest.approx(745.7, rel=1e-1)


def test_energy_flux():
    """Test EnergyFlux type."""
    energyf_type = energyflux.EnergyFlux()
    for unit in energyf_type.units:
        assert energyf_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = energyf_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = energyf_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in energyf_type.units:
            assert len(energyf_type.to_unit([1], other_unit, unit)) == 1
    assert energyf_type.to_unit([1], 'Btu/h-ft2', 'W/m2')[0] == pytest.approx(1 / 3.15459075, rel=1e-2)
    assert energyf_type.to_unit([1], 'kW/m2', 'W/m2')[0] == 0.001
    assert energyf_type.to_unit([1], 'kBtu/h-ft2', 'W/m2')[0] == pytest.approx(1 / 3154.59075, rel=1e-5)
    assert energyf_type.to_unit([1], 'W/ft2', 'W/m2')[0] == pytest.approx(1 / 10.7639, rel=1e-4)
    assert energyf_type.to_unit([1], 'met', 'W/m2')[0] == pytest.approx(1 / 58.2, rel=1e-4)
    assert energyf_type.to_unit([1], 'W/m2', 'Btu/h-ft2')[0] == pytest.approx(3.15459075, rel=1e-2)
    assert energyf_type.to_unit([1], 'W/m2', 'kW/m2')[0] == 1000
    assert energyf_type.to_unit([1], 'W/m2', 'kBtu/h-ft2')[0] == pytest.approx(3154.59075, rel=1e-1)
    assert energyf_type.to_unit([1], 'W/m2', 'W/ft2')[0] == pytest.approx(10.7639, rel=1e-1)
    assert energyf_type.to_unit([1], 'W/m2', 'met')[0] == pytest.approx(58.2, rel=1e-1)


def test_illuminance():
    """Test Illuminance type."""
    ill_type = illuminance.Illuminance()
    for unit in ill_type.units:
        assert ill_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = ill_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = ill_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in ill_type.units:
            assert len(ill_type.to_unit([1], other_unit, unit)) == 1
    assert ill_type.to_unit([1], 'fc', 'lux')[0] == pytest.approx(1 / 10.7639, rel=1e-3)
    assert ill_type.to_unit([1], 'lux', 'fc')[0] == pytest.approx(10.7639, rel=1e-1)


def test_luminance():
    """Test Luminance type."""
    lum_type = luminance.Luminance()
    for unit in lum_type.units:
        assert lum_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = lum_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = lum_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in lum_type.units:
            assert len(lum_type.to_unit([1], other_unit, unit)) == 1
    assert lum_type.to_unit([1], 'cd/ft2', 'cd/m2')[0] == pytest.approx(1 / 10.7639, rel=1e-3)
    assert lum_type.to_unit([1], 'cd/m2', 'cd/ft2')[0] == pytest.approx(10.7639, rel=1e-1)


def test_angle():
    """Test Angle type."""
    ang_type = angle.Angle()
    for unit in ang_type.units:
        assert ang_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = ang_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = ang_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in ang_type.units:
            assert len(ang_type.to_unit([1], other_unit, unit)) == 1
    assert ang_type.to_unit([1], 'radians', 'degrees')[0] == pytest.approx(PI / 180, rel=1e-3)
    assert ang_type.to_unit([1], 'degrees', 'radians')[0] == pytest.approx((1 / PI) * 180, rel=1e-1)


def test_mass():
    """Test Mass type."""
    mass_type = mass.Mass()
    for unit in mass_type.units:
        assert mass_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = mass_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = mass_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in mass_type.units:
            assert len(mass_type.to_unit([1], other_unit, unit)) == 1
    assert mass_type.to_unit([1], 'lb', 'kg')[0] == pytest.approx(2.20462, rel=1e-3)
    assert mass_type.to_unit([1], 'g', 'kg')[0] == 1000
    assert mass_type.to_unit([1], 'tonne', 'kg')[0] == 0.001
    assert mass_type.to_unit([1], 'ton', 'kg')[0] == pytest.approx(1 / 907.185, rel=1e-5)
    assert mass_type.to_unit([1], 'oz', 'kg')[0] == pytest.approx(35.274, rel=1e-1)
    assert mass_type.to_unit([1], 'kg', 'lb')[0] == pytest.approx(1 / 2.20462, rel=1e-3)
    assert mass_type.to_unit([1], 'kg', 'g')[0] == 0.001
    assert mass_type.to_unit([1], 'kg', 'tonne')[0] == 1000
    assert mass_type.to_unit([1], 'kg', 'ton')[0] == pytest.approx(907.185, rel=1e-1)
    assert mass_type.to_unit([1], 'kg', 'oz')[0] == pytest.approx(1 / 35.274, rel=1e-3)


def test_speed():
    """Test Speed type."""
    speed_type = speed.Speed()
    for unit in speed_type.units:
        assert speed_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = speed_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = speed_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in speed_type.units:
            assert len(speed_type.to_unit([1], other_unit, unit)) == 1
    assert speed_type.to_unit([1], 'mph', 'm/s')[0] == pytest.approx(2.23694, rel=1e-3)
    assert speed_type.to_unit([1], 'km/h', 'm/s')[0] == 3.6
    assert speed_type.to_unit([1], 'knot', 'm/s')[0] == pytest.approx(1.94384, rel=1e-3)
    assert speed_type.to_unit([1], 'ft/s', 'm/s')[0] == pytest.approx(3.28084, rel=1e-3)
    assert speed_type.to_unit([1], 'm/s', 'mph')[0] == pytest.approx(1 / 2.23694, rel=1e-3)
    assert speed_type.to_unit([1], 'm/s', 'km/h')[0] == 1 / 3.6
    assert speed_type.to_unit([1], 'm/s', 'knot')[0] == pytest.approx(1 / 1.94384, rel=1e-3)
    assert speed_type.to_unit([1], 'm/s', 'ft/s')[0] == pytest.approx(1 / 3.28084, rel=1e-3)


def test_volume_flow_rate():
    """Test VolumeFlowRate type."""
    vfr_type = volumeflowrate.VolumeFlowRate()
    for unit in vfr_type.units:
        assert vfr_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = vfr_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = vfr_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in vfr_type.units:
            assert len(vfr_type.to_unit([1], other_unit, unit)) == 1
    assert vfr_type.to_unit([1], 'ft3/s', 'm3/s')[0] == pytest.approx(35.3147, rel=1e-3)
    assert vfr_type.to_unit([1], 'L/s', 'm3/s')[0] == 1000
    assert vfr_type.to_unit([1], 'cfm', 'm3/s')[0] == pytest.approx(2118.88, rel=1e-1)
    assert vfr_type.to_unit([1], 'gpm', 'm3/s')[0] == pytest.approx(15850.3231, rel=1e-3)
    assert vfr_type.to_unit([1], 'mL/s', 'm3/s')[0] == 1000000
    assert vfr_type.to_unit([1], 'fl oz/s', 'm3/s')[0] == pytest.approx(33814, rel=1e-3)
    assert vfr_type.to_unit([1], 'm3/s', 'ft3/s')[0] == pytest.approx(1 / 35.3147, rel=1e-4)
    assert vfr_type.to_unit([1], 'm3/s', 'L/s')[0] == 0.001
    assert vfr_type.to_unit([1], 'm3/s', 'cfm')[0] == pytest.approx(1 / 2118.88, rel=1e-7)
    assert vfr_type.to_unit([1], 'm3/s', 'gpm')[0] == pytest.approx(1 / 15850.3231, rel=1e-8)
    assert vfr_type.to_unit([1], 'm3/s', 'mL/s')[0] == 0.000001
    assert vfr_type.to_unit([1], 'm3/s', 'fl oz/s')[0] == pytest.approx(1 / 33814, rel=1e-7)


def test_mass_flow_rate():
    """Test MassFlowRate type."""
    mfr_type = massflowrate.MassFlowRate()
    for unit in mfr_type.units:
        assert mfr_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = mfr_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = mfr_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in mfr_type.units:
            assert len(mfr_type.to_unit([1], other_unit, unit)) == 1
    assert mfr_type.to_unit([1], 'lb/s', 'kg/s')[0] == pytest.approx(2.2046, rel=1e-2)
    assert mfr_type.to_unit([1], 'g/s', 'kg/s')[0] == 1000
    assert mfr_type.to_unit([1], 'oz/s', 'kg/s')[0] == pytest.approx(35.274, rel=1e-2)
    assert mfr_type.to_unit([1], 'kg/s', 'lb/s')[0] == pytest.approx(1 / 2.2046, rel=1e-4)
    assert mfr_type.to_unit([1], 'kg/s', 'g/s')[0] == 0.001
    assert mfr_type.to_unit([1], 'kg/s', 'oz/s')[0] == pytest.approx(1 / 35.274, rel=1e-5)


def test_u_value():
    """Test UValue type."""
    uval_type = uvalue.UValue()
    for unit in uval_type.units:
        assert uval_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = uval_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = uval_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in uval_type.units:
            assert len(uval_type.to_unit([1], other_unit, unit)) == 1
    assert uval_type.to_unit([1], 'Btu/h-ft2-F', 'W/m2-K')[0] == pytest.approx(1 / 5.678263337, rel=1e-5)
    assert uval_type.to_unit([1], 'W/m2-K', 'Btu/h-ft2-F')[0] == pytest.approx(5.678263337, rel=1e-4)


def test_r_value():
    """Test RValue type."""
    rval_type = rvalue.RValue()
    for unit in rval_type.units:
        assert rval_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = rval_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = rval_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in rval_type.units:
            assert len(rval_type.to_unit([1], other_unit, unit)) == 1
    assert rval_type.to_unit([1], 'h-ft2-F/Btu', 'm2-K/W')[0] == pytest.approx(5.678263337, rel=1e-4)
    assert rval_type.to_unit([1], 'clo', 'm2-K/W')[0] == pytest.approx(1 / 0.155, rel=1e-2)
    assert rval_type.to_unit([1], 'm2-K/W', 'h-ft2-F/Btu')[0] == pytest.approx(1 / 5.678263337, rel=1e-5)
    assert rval_type.to_unit([1], 'm2-K/W', 'clo')[0] == pytest.approx(0.155, rel=1e-3)


def test_thermal_condition():
    """Test ThermalCondition type."""
    tc_type = thermalcondition.ThermalCondition()
    for unit in tc_type.units:
        assert tc_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = tc_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = tc_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in tc_type.units:
            assert len(tc_type.to_unit([1], other_unit, unit)) == 1
    assert tc_type.to_unit([1], 'PMV', 'condition')[0] == 1
    assert tc_type.to_unit([1], 'condition', 'PMV')[0] == 1


def test_specific_energy():
    """Test ThermalCondition type."""
    tc_type = specificenergy.SpecificEnergy()
    for unit in tc_type.units:
        assert tc_type.to_unit([1], unit, unit)[0] == pytest.approx(1, rel=1e-5)
        ip_vals, ip_u = tc_type.to_ip([1], unit)
        assert len(ip_vals) == 1
        si_vals, si_u = tc_type.to_si([1], unit)
        assert len(si_vals) == 1
        for other_unit in tc_type.units:
            assert len(tc_type.to_unit([1], other_unit, unit)) == 1
    assert tc_type.to_unit([1], 'kBtu/lb', 'kWh/kg')[0] == pytest.approx(1.54772, rel=1e-4)
    assert tc_type.to_unit([1], 'Btu/lb', 'kWh/kg')[0] == pytest.approx(1547.72, rel=1e-2)
    assert tc_type.to_unit([1], 'Wh/kg', 'kWh/kg')[0] == pytest.approx(1000, rel=1e-1)
    assert tc_type.to_unit([1], 'J/kg', 'kWh/kg')[0] == pytest.approx(3600000, rel=1e-1)
    assert tc_type.to_unit([1], 'kJ/kg', 'kWh/kg')[0] == pytest.approx(3600, rel=1e-1)
    assert tc_type.to_unit([1], 'kWh/kg', 'kBtu/lb')[0] == pytest.approx(0.646111699, rel=1e-5)
    assert tc_type.to_unit([1], 'kWh/kg', 'Btu/lb')[0] == pytest.approx(0.00064611169979, rel=1e-7)
    assert tc_type.to_unit([1], 'kWh/kg', 'Wh/kg')[0] == pytest.approx(0.001, rel=1e-3)
    assert tc_type.to_unit([1], 'kWh/kg', 'J/kg')[0] == pytest.approx(2.7777777777777776e-07, rel=1e-9)
    assert tc_type.to_unit([1], 'kWh/kg', 'kJ/kg')[0] == pytest.approx(0.0002777777777777778, rel=1e-7)



# coding=utf-8
from pytest import approx
import os
from ladybug.location import Location
from ladybug.dt import Date
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.designday import DesignDay
from ladybug.ddy import DDY


def test_import_ddy():
    """Test import standard ddy."""
    relative_path = './tests/fixtures/ddy/chicago.ddy'
    abs_path = os.path.abspath(relative_path)

    ddy_rel = DDY.from_ddy_file(relative_path)
    ddy = DDY.from_ddy_file(abs_path)

    # Test imports don't break
    assert ddy.file_path == abs_path
    assert ddy_rel.file_path == os.path.normpath(relative_path)


def test_dict_methods():
    """Test dict methods for the DDY object."""
    relative_path = './tests/fixtures/ddy/chicago.ddy'
    ddy = DDY.from_ddy_file(relative_path)

    ddy_dict = ddy.to_dict()
    reconstructed_ddy = DDY.from_dict(ddy_dict)
    assert ddy_dict == reconstructed_ddy.to_dict()
    for dday1, dday2 in zip(ddy.design_days, reconstructed_ddy.design_days):
        assert dday1 == dday2


def test_ddy_from_design_day():
    """Test ddy from design day method."""
    relative_path = './tests/fixtures/ddy/chicago_monthly.ddy'
    ddy = DDY.from_ddy_file(relative_path)
    new_ddy = DDY.from_design_day(ddy.design_days[0])
    assert ddy.location == new_ddy.location
    assert ddy.design_days[0] == new_ddy.design_days[0]


def test_write_ddy():
    """Test write ddy."""
    relative_path = './tests/fixtures/ddy/chicago.ddy'
    ddy = DDY.from_ddy_file(relative_path)
    new_file_path = './tests/fixtures/ddy/chicago_edited.ddy'
    ddy.save(new_file_path)


def test_standard_ddy_properties():
    """Test properties of a standard ddy."""
    relative_path = './tests/fixtures/ddy/tokyo.ddy'

    ddy = DDY.from_ddy_file(relative_path)

    # Test accuracy of import
    assert ddy.location.city == 'TOKYO HYAKURI_JPN Design_Conditions'
    assert ddy.location.latitude == approx(36.18, rel=1e-1)
    assert ddy.location.longitude == approx(140.42, rel=1e-1)
    assert ddy.location.time_zone == 9
    assert ddy.location.elevation == 35

    assert len(ddy.design_days) == 18
    for des_day in ddy.design_days:
        assert isinstance(des_day, DesignDay)
    assert len(ddy.filter_by_keyword('.4%')) == 4
    assert len(ddy.filter_by_keyword('99.6%')) == 3


def test_duplicate():
    """Test duplicate method for the DDY object."""
    relative_path = './tests/fixtures/ddy/chicago_monthly.ddy'
    ddy = DDY.from_ddy_file(relative_path)
    ddy_dup = ddy.duplicate()

    assert ddy is ddy
    assert ddy is not ddy_dup
    assert ddy == ddy_dup
    ddy_dup[0].dry_bulb_condition.dry_bulb_max = 40
    assert ddy != ddy_dup


def test_duplicate_design_day():
    """Test duplicate method for the DesignDay object."""
    relative_path = './tests/fixtures/ddy/chicago_monthly.ddy'
    ddy = DDY.from_ddy_file(relative_path)

    des_day = ddy[0]
    des_day_dup = des_day.duplicate()

    assert des_day is des_day
    assert des_day is not des_day_dup
    assert des_day == des_day_dup
    des_day_dup.dry_bulb_condition.dry_bulb_max = 40
    assert des_day != des_day_dup


def test_monthly_ddy_properties():
    """Test properties of a monthly ddy."""
    relative_path = './tests/fixtures/ddy/chicago_monthly.ddy'
    ddy = DDY.from_ddy_file(relative_path)

    # Test accuracy of import
    assert ddy.location.city == 'Chicago Ohare Intl Ap'
    assert ddy.location.latitude == approx(41.96, rel=1e-1)
    assert ddy.location.longitude == approx(-87.92, rel=1e-1)
    assert ddy.location.time_zone == -6
    assert ddy.location.elevation == 201

    assert len(ddy.design_days) == 12
    for des_day in ddy.design_days:
        assert isinstance(des_day, DesignDay)
        assert des_day.day_type == 'SummerDesignDay'


def test_design_day_from_properties():
    """Test hourly data properties of a standard ddy."""
    location = Location('Test City', '-', 'USA', 34.20, -118.35, -8, 226)
    a_period = AnalysisPeriod(12, 21, 0, 12, 21, 23)
    date = Date(12, 21)
    des_day = DesignDay.from_design_day_properties(
        'Test Day', 'WinterDesignDay', location, date, 3.9, 0,
        'Wetbulb', 3.9, 98639, 0.8, 330, 'ASHRAEClearSky', [0])
    assert des_day.location == location
    new_period = des_day.analysis_period
    assert new_period.st_month == a_period.st_month
    assert new_period.st_day == a_period.st_day
    assert new_period.st_hour == a_period.st_hour
    assert new_period.end_month == a_period.end_month
    assert new_period.end_day == a_period.end_day
    assert new_period.end_hour == a_period.end_hour


def test_design_day_hourly_data():
    """Test hourly data properties of a standard ddy."""
    location = Location('Test City', '-', 'USA', 34.20, -118.35, -8, 226)
    date = Date(8, 21)
    des_day = DesignDay.from_design_day_properties(
        'Test Day', 'SummerDesignDay', location, date, 36.8, 13.2,
        'Wetbulb', 20.5, 98639, 3.9, 170, 'ASHRAETau', [0.436, 2.106])
    # dry bulb values
    db_data_collect = des_day.hourly_dry_bulb
    assert db_data_collect[5] == approx(23.6, rel=1e-1)
    assert db_data_collect[14] == approx(36.8, rel=1e-1)

    # dew point values
    dpt_data_collect = des_day.hourly_dew_point
    assert dpt_data_collect[0] == approx(11.296, rel=1e-1)
    assert dpt_data_collect[-1] == approx(11.296, rel=1e-1)

    # relative humidity values
    rh_data_collect = des_day.hourly_relative_humidity
    assert rh_data_collect[5] == approx(45.896, rel=1e-1)
    assert rh_data_collect[14] == approx(21.508, rel=1e-1)

    # barometric pressure values
    bp_data_collect = des_day.hourly_barometric_pressure
    assert bp_data_collect[0] == approx(98639, rel=1e-1)
    assert bp_data_collect[-1] == approx(98639, rel=1e-1)

    # wind speed values
    ws_data_collect = des_day.hourly_wind_speed
    assert -0.1 < ws_data_collect[0] - 3.9 < 0.1
    assert -0.1 < ws_data_collect[-1] - 3.9 < 0.1

    # wind direction values
    wd_data_collect = des_day.hourly_wind_direction
    assert wd_data_collect[0] == approx(170, rel=1e-1)
    assert wd_data_collect[-1] == approx(170, rel=1e-1)

    # radiation values
    direct_normal_rad, diffuse_horizontal_rad, global_horizontal_rad = \
        des_day.hourly_solar_radiation
    assert direct_normal_rad[0] == 0
    assert direct_normal_rad[11] == approx(891.46, rel=1e-1)
    assert diffuse_horizontal_rad[0] == 0
    assert diffuse_horizontal_rad[11] == approx(165.32, rel=1e-1)
    assert global_horizontal_rad[0] == 0
    assert global_horizontal_rad[11] == approx(985.05, rel=1e-1)

    # sky cover values
    sc_data_collect = des_day.hourly_sky_cover

    # sky cover values
    hi_data_collect = des_day.hourly_horizontal_infrared



# coding=utf-8
"""Generic data type."""
from __future__ import division

from .base import DataTypeBase


class GenericType(DataTypeBase):
    """Type for any data type that is not currently implemented.

    Args:
        name: A name for the data type as a string.
        unit: A unit for the data type as a string.
        min: Optional lower limit for the data type, values below which should be
            physically or mathematically impossible. (Default: -inf)
        max: Optional upper limit for the data type, values above which should be
            physically or mathematically impossible. (Default: +inf)
        abbreviation: An optional abbreviation for the data type as text.
        unit_descr: An optional dictionary describing categories that the numerical
            values of the units relate to. For example:
            {-1: 'Cold', 0: 'Neutral', +1: 'Hot'}
            {0: 'False', 1: 'True'}
        point_in_time: Boolean to note whether the data type represents conditions
            at a single instant in time (True) as opposed to being an average or
            accumulation over time (False) when it is found in hourly lists of data.
            (Default: True)
        cumulative: Boolean to tell whether the data type can be cumulative when it
            is represented over time (True) or it can only be averaged over time
            to be meaningful (False). Note that cumulative cannot be True
            when point_in_time is also True. (Default: False)
    """
    def __init__(self, name, unit, min=float('-inf'), max=float('+inf'),
                 abbreviation=None, unit_descr=None, point_in_time=True,
                 cumulative=False):
        """Initialize Generic Type.
        """
        assert isinstance(name, str), 'name must be a string. Got {}.'.format(type(name))
        assert isinstance(unit, str), 'unit must be a string. Got {}.'.format(type(unit))
        assert isinstance(min, (float, int)), 'min must be a number. ' \
            'Got {}.'.format(type(min))
        assert isinstance(max, (float, int)), 'max must be a number. ' \
            'Got {}.'.format(type(max))
        if abbreviation is not None:
            assert isinstance(abbreviation, str), 'abbreviation must be a ' \
                'string. Got {}.'.format(type(abbreviation))
        if unit_descr is not None:
            assert isinstance(unit_descr, dict), 'unit_descr must be a ' \
                'dictionary. Got {}.'.format(type(unit_descr))
        assert isinstance(point_in_time, bool), 'point_in_time must be a ' \
            'boolean. Got {}.'.format(type(point_in_time))
        assert isinstance(cumulative, bool), 'cumulative must be a ' \
            'boolean. Got {}.'.format(type(cumulative))
        if point_in_time:
            assert not cumulative, 'cumulative cannot be True when ' \
                'point_in_time is also True.'

        self._name = name
        self._units = [unit]
        self._min = min
        self._max = max
        self._abbreviation = abbreviation if abbreviation is not None else name
        self._unit_descr = unit_descr
        self._point_in_time = point_in_time
        self._cumulative = cumulative

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        return values, from_unit

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        return values, from_unit



# coding=utf-8
"""Generic data type."""
from __future__ import division

from .base import DataTypeBase


class MassFlowRate(DataTypeBase):
    """MassFlowRate
    """
    _units = ('kg/s', 'lb/s', 'g/s', 'oz/s')
    _si_units = ('kg/s', 'g/s')
    _ip_units = ('lb/s', 'oz/s')
    _min = 0
    _abbreviation = 'dm/dt'

    def _kg_s_to_lb_s(self, value):
        return value * 2.2046

    def _kg_s_to_g_s(self, value):
        return value * 1000.

    def _kg_s_to_oz_s(self, value):
        return value * 35.274

    def _lb_s_to_kg_s(self, value):
        return value / 2.20462

    def _g_s_to_kg_s(self, value):
        return value / 1000.

    def _oz_s_to_kg_s(self, value):
        return value / 35.274

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('kg/s', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'g/s':
            return self.to_unit(values, 'oz/s', from_unit), 'oz/s'
        else:
            return self.to_unit(values, 'lb/s', from_unit), 'lb/s'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'oz/s':
            return self.to_unit(values, 'g/s', from_unit), 'g/s'
        else:
            return self.to_unit(values, 'kg/s', from_unit), 'kg/s'



# coding=utf-8
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.dt import DateTime

from datetime import timedelta
import sys
if (sys.version_info >= (3, 0)):
    xrange = range


def test_default_values():
    """Test the default values."""
    ap = AnalysisPeriod()
    assert ap.st_hour == 0
    assert ap.end_hour == 23
    assert ap.st_month == 1
    assert ap.end_month == 12
    assert ap.st_day == 1
    assert ap.end_day == 31
    assert ap.timestep == 1
    assert not ap.is_leap_year
    assert len(ap.datetimes) == 8760


def test_from_string():
    """Test creating analysis priod from a string."""
    ap_string = '2/21 to 3/22 between 5 to 17 @1'
    ap = AnalysisPeriod.from_analysis_period(ap_string)
    ap = AnalysisPeriod.from_string(ap_string)
    assert ap.st_hour == 5
    assert ap.end_hour == 17
    assert ap.st_month == 2
    assert ap.end_month == 3
    assert ap.st_day == 21
    assert ap.end_day == 22
    assert ap.timestep == 1


def test_is_annual():
    """Test the is_annual property."""
    annual_ap = AnalysisPeriod()
    not_annual_ap = AnalysisPeriod(end_month=2)
    assert annual_ap.is_annual
    assert not not_annual_ap.is_annual


def test_to_from_dict():
    """Test the dict methods of the AnalysisPeriod."""
    ap = AnalysisPeriod(st_month=2)
    ap_dict = ap.to_dict()
    rebuilt_ap = AnalysisPeriod.from_dict(ap_dict)
    assert rebuilt_ap.st_month == 2
    assert rebuilt_ap.to_dict() == ap_dict


def test_include_last_hour():
    """Test that analysis period includes the last hour.

    see:
    https://github.com/ladybug-tools/ladybug-dynamo/issues/12#issuecomment-224125154
    """
    ap = AnalysisPeriod(7, 21, 9, 7, 21, 15, 4)
    last_hour = ap.datetimes[-1]
    assert last_hour.hour == 15
    assert last_hour.minute == 0
    assert len(ap) == 25


def test_default_values_leap_year():
    """Test the default values."""
    ap = AnalysisPeriod(is_leap_year=True)
    assert ap.st_hour == 0
    assert ap.end_hour == 23
    assert ap.st_month == 1
    assert ap.end_month == 12
    assert ap.st_day == 1
    assert ap.end_day == 31
    assert ap.timestep == 1
    assert len(ap.datetimes) == 8784


def test_from_string_leap_year():
    """Test creating analysis priod from a string."""
    ap_string = '2/21 to 3/22 between 5 to 17 @1*'
    ap = AnalysisPeriod.from_string(ap_string)
    assert ap.st_hour == 5
    assert ap.end_hour == 17
    assert ap.st_month == 2
    assert ap.end_month == 3
    assert ap.st_day == 21
    assert ap.end_day == 22
    assert ap.timestep == 1


def test_is_annual_leap_year():
    annual_ap = AnalysisPeriod(is_leap_year=True)
    not_annual_ap = AnalysisPeriod(end_month=2, is_leap_year=True)
    assert annual_ap.is_annual
    assert not not_annual_ap.is_annual


def test_moys_hoys_hoys_int():
    """Test the moys, hoys, and hoys_int properties."""
    ap = AnalysisPeriod(timestep=4)
    assert ap.moys[1] == 15
    assert ap.hoys[1] == 0.25
    assert ap.hoys_int[1] == 0
    assert isinstance(ap.hoys[1], float)
    assert isinstance(ap.hoys_int[1], int)
    assert len(ap.moys) == len(ap.hoys) == len(ap.hoys_int)


def test_doys_int():
    """Test the doys_int property."""
    ap = AnalysisPeriod()
    ap_2 = AnalysisPeriod(st_month=2)
    assert ap.doys_int == list(xrange(1, 366))
    assert ap_2.doys_int == list(xrange(32, 366))


def test_months_int():
    """Test the months_int property."""
    ap = AnalysisPeriod()
    ap_2 = AnalysisPeriod(st_month=2)
    assert ap.months_int == list(xrange(1, 13))
    assert ap_2.months_int == list(xrange(2, 13))


def test_months_per_hour():
    """Test the months_per_hour property."""
    ap = AnalysisPeriod()
    ap_2 = AnalysisPeriod(st_month=2)
    ap_3 = AnalysisPeriod(st_hour=5)
    assert len(ap.months_per_hour) == 12 * 24
    assert len(ap_2.months_per_hour) == 11 * 24
    assert len(ap_3.months_per_hour) == 12 * 19


def test_minute_intervals():
    """Test the minute_intervals property."""
    ap = AnalysisPeriod()
    ap_2 = AnalysisPeriod(timestep=6)
    assert ap.minute_intervals == timedelta(minutes=60)
    assert ap_2.minute_intervals == timedelta(minutes=10)


def test_is_reversed():
    """Test the is_reversed property."""
    ap = AnalysisPeriod()
    ap_2 = AnalysisPeriod(st_month=6, end_month=2)
    assert not ap.is_reversed
    assert ap_2.is_reversed


def test_is_overnight():
    """Test the is_overnight property."""
    ap = AnalysisPeriod()
    ap_2 = AnalysisPeriod(st_hour=20, end_hour=5)
    assert not ap.is_overnight
    assert ap_2.is_overnight


def test_is_time_included():
    """Test the is_time_included method."""
    ap = AnalysisPeriod(st_month=2, end_month=6)
    assert ap.is_time_included(DateTime(5, 18, 17))
    assert not ap.is_time_included(DateTime(1, 10, 12))


def test_duplicate():
    """Test the duplicate method."""
    ap = AnalysisPeriod()
    ap_2 = ap.duplicate()
    assert ap_2 == ap


def test_0_end_hour():
    """Test hour 0 for end hour."""
    params = [1, 1, 1, 1, 2, 0]
    ap = AnalysisPeriod(*params)
    assert ap.end_hour == 0
    assert len(ap) == 24

def test_equality():
    ap_one = AnalysisPeriod()
    ap_one_duplicate = AnalysisPeriod()
    ap_two = AnalysisPeriod(st_month=2, end_month=6)

    assert ap_one is ap_one
    assert ap_one == ap_one_duplicate
    assert ap_one is not ap_two
    assert hash(ap_one) == hash(ap_one_duplicate)
    assert hash(ap_one) != hash(ap_two)



# coding=utf-8
"""Generic data type."""
from __future__ import division

from .base import DataTypeBase
from .energyflux import EnergyFlux


class Power(DataTypeBase):
    """Power
    """
    _units = ('W', 'Btu/h', 'kW', 'kBtu/h', 'TR', 'hp')
    _si_units = ('kW', 'W')
    _ip_units = ('Btu/h', 'kBtu/h', 'TR', 'hp')
    _abbreviation = 'Q'
    _point_in_time = False
    _normalized_type = EnergyFlux

    def _W_to_Btu_h(self, value):
        return value * 3.41214

    def _W_to_kW(self, value):
        return value / 1000.

    def _W_to_kBtu_h(self, value):
        return value * 0.00341214

    def _W_to_TR(self, value):
        return value / 3516.85

    def _W_to_hp(self, value):
        return value / 745.7

    def _Btu_h_to_W(self, value):
        return value / 3.41214

    def _kW_to_W(self, value):
        return value * 1000.

    def _kBtu_h_to_W(self, value):
        return value / 0.00341214

    def _TR_to_W(self, value):
        return value * 3516.85

    def _hp_to_W(self, value):
        return value * 745.7

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('W', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'kW':
            return self.to_unit(values, 'kBtu/h', from_unit), 'kBtu/h'
        else:
            return self.to_unit(values, 'Btu/h', from_unit), 'Btu/h'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'kBtu/h':
            return self.to_unit(values, 'kW', from_unit), 'kW'
        else:
            return self.to_unit(values, 'W', from_unit), 'W'


class ActivityLevel(Power):
    _abbreviation = 'Activity'
    _min = 0



# coding=utf-8
"""Energy flux data type."""
from __future__ import division

from .base import DataTypeBase


class EnergyFlux(DataTypeBase):
    """Energy Flux
    """
    _units = ('W/m2', 'Btu/h-ft2', 'kW/m2', 'kBtu/h-ft2', 'W/ft2', 'met')
    _si_units = ('W/m2', 'kW/m2')
    _ip_units = ('Btu/h-ft2', 'kBtu/h-ft2')
    _abbreviation = 'J'
    _point_in_time = False

    def _W_m2_to_Btu_hft2(self, value):
        return value / 3.15459075

    def _W_m2_to_kW_m2(self, value):
        return value / 1000.

    def _W_m2_to_kBtu_hft2(self, value):
        return value / 3154.59075

    def _W_m2_to_W_ft2(self, value):
        return value / 10.7639

    def _W_m2_to_met(self, value):
        return value / 58.2

    def _Btu_hft2_to_W_m2(self, value):
        return value * 3.15459075

    def _kW_m2_to_W_m2(self, value):
        return value * 1000.

    def _kBtu_hft2_to_W_m2(self, value):
        return value * 3154.59075

    def _W_ft2_to_W_m2(self, value):
        return value * 10.7639

    def _met_to_W_m2(self, value):
        return value * 58.2

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('W/m2', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units or from_unit == 'met':
            return values, from_unit
        elif from_unit == 'kW/m2':
            return self.to_unit(values, 'kBtu/h-ft2', from_unit), 'kBtu/h-ft2'
        else:
            return self.to_unit(values, 'Btu/h-ft2', from_unit), 'Btu/h-ft2'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units or from_unit == 'met':
            return values, from_unit
        elif from_unit == 'kBtu/h-ft2':
            return self.to_unit(values, 'kW/m2', from_unit), 'kW/m2'
        else:
            return self.to_unit(values, 'W/m2', from_unit), 'W/m2'


class MetabolicRate(EnergyFlux):
    _min = 0
    _abbreviation = 'MetR'


class EffectiveRadiantField(EnergyFlux):
    _abbreviation = 'ERF'


class Irradiance(EnergyFlux):
    _min = 0
    _abbreviation = 'Qsolar'


class GlobalHorizontalIrradiance(Irradiance):
    _abbreviation = 'GHIr'


class DirectNormalIrradiance(Irradiance):
    _abbreviation = 'DNIr'


class DiffuseHorizontalIrradiance(Irradiance):
    _abbreviation = 'DHIr'


class DirectHorizontalIrradiance(Irradiance):
    _abbreviation = 'DHIr'


class HorizontalInfraredRadiationIntensity(Irradiance):
    _abbreviation = 'HIr'
    _point_in_time = True



# coding=utf-8
from __future__ import division

from .color import Color, Colorset, ColorRange

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.mesh import Mesh3D
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.mesh import Mesh2D

try:
    from collections.abc import Iterable  # python < 3.7
except ImportError:
    from collections import Iterable  # python >= 3.8
import sys
if (sys.version_info > (3, 0)):  # python 3
    xrange = range


class Legend(object):
    """Ladybug legend used to get legend geometry, legend text, generate colors, etc.

    Args:
        values: A List or Tuple of numerical values that will be used to
            generate the legend and colors.
        legend_parameters: An Optional LegendParameter object to override
            default parameters of the legend.

    Properties:
        * legend_parameters
        * values
        * value_colors
        * title
        * title_location
        * title_location_2d
        * segment_text
        * segment_text_location
        * segment_text_location_2d
        * segment_mesh
        * segment_mesh_2d
        * color_range
        * segment_numbers
        * segment_colors
        * segment_length
        * is_min_default
        * is_max_default

    Usage:

    .. code-block:: python

        1.
        data = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        legend = Legend(data, LegendParameters(segment_count=6))
        print(legend.segment_text)
        print(legend.segment_mesh)
        print(legend.segment_colors)

        >> ['0.00', '1.80', '3.60', '5.40', '7.20', '9.00']
        >> Mesh3D (6 faces) (14 vertices)
        >> ((R:75, G:107, B:169), (R:159, G:189, B:238), (R:224, G:229, B:145),
            (R:247, G:200, B:53), (R:234, G:113, B:0), (R:234, G:38, B:0))

        2.
        data = [100, 300, 500, 1000, 2000, 3000]
        leg_colors = [Color(0, 0, 255), Color(0, 255, 0), Color(255, 0, 0)]
        legend_par = LegendParametersCategorized([300, 2000], leg_colors)
        legend_par.category_names = ['low', 'desired', 'too much']
        legend_par.continuous_colors = False  # get data in only 3 colors
        legend = Legend(data, legend_par)
        print(legend.segment_text)
        print(legend.segment_mesh)
        print(legend.segment_colors)
        print(legend.value_colors)
        >> ['low', 'desired', 'too much']
        >> Mesh3D (3 faces) (8 vertices)
        >> ((R:0, G:0, B:255), (R:0, G:255, B:0), (R:255, G:0, B:0))
        >> ((R:0, G:0, B:255), (R:0, G:255, B:0), (R:0, G:255, B:0),
            (R:0, G:255, B:0), (R:0, G:255, B:0), (R:255, G:0, B:0))

        3.
        data = [-0.5, 0, 0.5]
        legend_par = LegendParameters(min=-1, max=1, segment_count=3)
        legend_par.ordinal_dictionary = {-3: 'Cold', -2: 'Cool', -1: 'Slightly Cool',
                                         0: 'Neutral',
                                         1: 'Slightly Warm', 2: 'Warm', 3: 'Hot'}
        legend = Legend(data, legend_par)
        print(legend.segment_text)
        print(legend.segment_mesh)
        print(legend.segment_colors)
        legend.legend_parameters.segment_count = 5
        print(legend.segment_text)
        legend.legend_parameters.min = -2
        legend.legend_parameters.max = 2
        print(legend.segment_text)
        >> ['Slightly Cool', 'Neutral', 'Slightly Warm']
        >> Mesh3D (3 faces) (8 vertices)
        >> ((R:75, G:107, B:169), (R:249, G:235, B:89), (R:234, G:38, B:0))
        >> ['Slightly Cool', '', 'Neutral', '', 'Slightly Warm']
        >> ['Cool', 'Slightly Cool', 'Neutral', 'Slightly Warm', 'Warm']
    """
    __slots__ = ('_values', '_legend_par', '_is_min_default', '_is_max_default')

    def __init__(self, values, legend_parameters=None):
        """Initialize Ladybug Legend.
        """
        # check the inputs
        assert isinstance(values, Iterable) \
            and not isinstance(values, (str, dict, bytes, bytearray)), \
            'values should be a list or tuple. Got {}'.format(type(values))
        self._values = values
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), \
                'Expected LegendParameters. Got {}.'.format(type(legend_parameters))
            self._legend_par = legend_parameters.duplicate()
        else:
            self._legend_par = LegendParameters()

        # set default min, max and segment count (if min == max)
        self._is_min_default = False
        self._is_max_default = False
        if self._legend_par.min is None:
            self._legend_par.min = min(values)
            self._is_min_default = True
        if self._legend_par.max is None:
            self._legend_par.max = max(values)
            self._is_max_default = True
        if self._legend_par.min == self._legend_par.max and not \
                isinstance(self._legend_par, LegendParametersCategorized) and \
                self._legend_par.is_segment_count_default:
            self._legend_par.segment_count = 1

        # set the default segment width if the legend is horizontal
        if not self._legend_par.vertical and self._legend_par.is_segment_width_default:
            self._legend_par.segment_width = self._legend_par.text_height * \
                (len(str(int(self._legend_par.max))) +
                 self._legend_par.decimal_count + 2)
            self._legend_par._is_segment_width_default = True  # it's been auto-assigned

    @classmethod
    def from_dict(cls, data):
        """Create a legend from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "values": [0, 10],
            "legend_parameters": None
            }
        """
        legend_parameters = None
        if 'legend_parameters' in data and data['legend_parameters'] is not None:
            if data['legend_parameters']['type'] == 'LegendParametersCategorized':
                legend_parameters = LegendParametersCategorized.from_dict(
                    data['legend_parameters'])
            else:
                legend_parameters = LegendParameters.from_dict(data['legend_parameters'])

        legend = cls(data['values'], legend_parameters)
        legend._is_min_default = data['is_min_default'] if 'is_min_default' in data \
            else False
        legend._is_max_default = data['is_max_default'] if 'is_max_default' in data \
            else False
        return legend

    @property
    def legend_parameters(self):
        """The legend parameters assigned to this legend."""
        return self._legend_par

    @property
    def values(self):
        """The data set assigned to the legend."""
        return self._values

    @property
    def value_colors(self):
        """A List of colors associated with the assigned values."""
        _color_range = self.color_range
        return tuple(_color_range.color(val) for val in self.values)

    @property
    def title(self):
        """A text string for the title of the legend."""
        return self.legend_parameters.title

    @property
    def title_location(self):
        """A Plane for the location and orientation of the legend title."""
        _base_pl = self.legend_parameters.base_plane
        _title_pt = self._title_point_2d()
        return Plane(_base_pl.n, _base_pl.xy_to_xyz(_title_pt), _base_pl.x)

    @property
    def title_location_2d(self):
        """A Point2D for the location of the title.

        Useful for output to 2D interfaces.
        """
        _base_o = self.legend_parameters.base_plane.o
        _title_pt = self._title_point_2d()
        return Point2D(_title_pt.x + _base_o.x, _title_pt.y + _base_o.y)

    @property
    def segment_text(self):
        """A list of text strings for the segment labels of the legend."""
        _l_par = self.legend_parameters
        if isinstance(_l_par, LegendParametersCategorized):
            return _l_par.category_names
        else:
            if _l_par.ordinal_dictionary is None:
                format_str = '%.{}f'.format(_l_par.decimal_count)
                seg_txt = [format_str % x for x in self.segment_numbers]
                if _l_par.include_larger_smaller:
                    seg_txt[0] = '<' + seg_txt[0]
                    seg_txt[-1] = '>' + seg_txt[-1]
                return seg_txt
            else:
                seg_txt = []
                for x in self.segment_numbers:
                    try:
                        seg_txt.append(_l_par.ordinal_dictionary[x])
                    except KeyError:
                        seg_txt.append('')
            return seg_txt

    @property
    def segment_text_location(self):
        """A list of Plane objects for the location of the legend segment text."""
        _base_pl = self.legend_parameters.base_plane
        _pt_2d = self._segment_point_2d()
        return [Plane(_base_pl.n, _base_pl.xy_to_xyz(pt), _base_pl.x) for pt in _pt_2d]

    @property
    def segment_text_location_2d(self):
        """A list of Point2D for the location of the legend segment text.

        Useful for output to 2D interfaces.
        """
        _base_o = self.legend_parameters.base_plane.o
        _pt_2d = self._segment_point_2d()
        return [Point2D(pt.x + _base_o.x, pt.y + _base_o.y) for pt in _pt_2d]

    @property
    def segment_mesh(self):
        """A Ladybug Mesh3D for the legend colors."""
        _mesh_2d = self._segment_mesh_2d()
        return Mesh3D.from_mesh2d(_mesh_2d, self.legend_parameters.base_plane)

    @property
    def segment_mesh_2d(self):
        """A Ladybug Mesh2D for the legend colors."""
        _o = self.legend_parameters.base_plane.o
        return self._segment_mesh_2d(Point2D(_o.x, _o.y))

    @property
    def color_range(self):
        """The color range associated with this legend."""
        _l_par = self.legend_parameters
        if isinstance(_l_par, LegendParametersCategorized):
            return ColorRange(_l_par.colors, _l_par.domain, _l_par.continuous_colors)
        else:
            return ColorRange(_l_par.colors, (_l_par.min, _l_par.max))

    @property
    def segment_numbers(self):
        """Get a list of numbers along a linear scale from the min to max."""
        _l_par = self.legend_parameters
        try:
            _seg_stp = (_l_par.max - _l_par.min) / (_l_par.segment_count - 1)
        except ZeroDivisionError:
            _seg_stp = 0
        return tuple(_l_par.min + i * _seg_stp
                     for i in xrange(_l_par.segment_count))

    @property
    def segment_colors(self):
        """A list of colors associated with the legend segments."""
        if isinstance(self.legend_parameters, LegendParametersCategorized):
            return self.legend_parameters.colors
        _color_range = self.color_range
        return tuple(_color_range.color(val) for val in self.segment_numbers)

    @property
    def segment_length(self):
        """An integer for the number of segment lengths in the legend."""
        _l_par = self.legend_parameters
        return _l_par.segment_count if not _l_par.continuous_legend else \
            _l_par.segment_count - 1

    @property
    def is_min_default(self):
        """Boolean noting whether the min is default.

        Useful when deciding whether to override the legend to properly display an
        ordinal dictionary.
        """
        return self._is_min_default

    @property
    def is_max_default(self):
        """Boolean noting whether the max is default.

        Useful when deciding whether to override the legend to properly display an
        ordinal dictionary.
        """
        return self._is_max_default

    def duplicate(self):
        """Return a copy of the current legend."""
        return self.__copy__()

    def to_dict(self):
        """Get legend as a dictionary."""
        return {
            'values': self.values,
            'legend_parameters': self.legend_parameters.to_dict(),
            'is_min_default': self.is_min_default,
            'is_max_default': self.is_max_default,
            'type': 'Legend'
        }

    def _title_point_2d(self):
        """Point2D for the title in the 2D space of the legend."""
        _l_par = self.legend_parameters
        if _l_par.vertical:
            offset = 0.5 if self.legend_parameters.continuous_legend else 0.25
            base = Point2D(0, _l_par.segment_height * (self.segment_length + offset))
        else:
            base = Point2D(-_l_par.segment_width * self.segment_length,
                           _l_par.segment_height * 1.25)
        ln_break_count = self.legend_parameters.title.count('\n')
        if ln_break_count != 0:  # offset the text so that it's not over the legend
            offset = ln_break_count * self.legend_parameters.text_height * 1.5
            return Point2D(base.x, base.y + offset)
        return base

    def _segment_point_2d(self):
        """Point2D for the segment text in the 2D space of the legend."""
        _l_par = self.legend_parameters
        if _l_par.vertical:  # vertical
            _pt_2d = tuple(
                Point2D(_l_par.segment_width + _l_par.text_height * 0.25, i)
                for i in Legend._frange(0, _l_par.segment_height *
                                        _l_par.segment_count,
                                        _l_par.segment_height))
        else:  # horizontal
            _start_val = -_l_par.segment_width * self.segment_length
            _pt_2d = tuple(
                Point2D(_start_val + i, -_l_par.text_height * 1.25)
                for i in Legend._frange(0, _l_par.segment_width *
                                        _l_par.segment_count,
                                        _l_par.segment_width))
        return _pt_2d

    def _segment_mesh_2d(self, base_pt=Point2D(0, 0)):
        """Mesh2D for the segments in the 2D space of the legend."""
        # get general properties
        _l_par = self.legend_parameters
        n_seg = self.segment_length
        # create the 2D mesh of the legend
        if _l_par.vertical:
            mesh2d = Mesh2D.from_grid(
                base_pt, 1, n_seg, _l_par.segment_width, _l_par.segment_height)
        else:
            _base_pt = Point2D(base_pt.x - _l_par.segment_width * n_seg, base_pt.y)
            mesh2d = Mesh2D.from_grid(
                _base_pt, n_seg, 1, _l_par.segment_width, _l_par.segment_height)
        # add colors to the mesh
        _seg_colors = self.segment_colors
        if not _l_par.continuous_legend:
            mesh2d.colors = _seg_colors
        else:
            if _l_par.vertical:
                mesh2d.colors = _seg_colors + _seg_colors
            else:
                mesh2d.colors = tuple(col for col in _seg_colors for i in (0, 1))
        return mesh2d

    @staticmethod
    def _frange(start, stop, step):
        """Range function capable of yielding float values."""
        while start < stop:
            yield start
            start += step

    def __copy__(self):
        _leg = Legend(self.values, self.legend_parameters)
        _leg._is_min_default = self._is_min_default
        _leg._is_max_default = self._is_max_default
        return _leg

    def __len__(self):
        """Return length of values on the object."""
        return len(self._values)

    def __getitem__(self, key):
        """Return one of the values."""
        return self._values[key]

    def __iter__(self):
        """Iterate through the values."""
        return iter(self._values)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Legend representation."""
        return 'Ladybug Legend ({} values)'.format(len(self))


class LegendParameters(object):
    """Ladybug legend parameters used to customize legends.

    All properties of LegendParameters are set-able (except the
    is_someproperty_default ones).

    Args:
        min: A number to set the lower boundary of the legend. If None, the
            minimum of the values associated with the legend will be used.
        max: A number to set the upper boundary of the legend. If None, the
            maximum of the values associated with the legend will be used.
        segment_count: An interger representing the number of steps between
            the high and low boundary of the legend. The default is set to 11
            and any custom values input in here should always be greater than or
            equal to 2.
        colors: An list of color objects. Default is Ladybug's original colorset.
        title: Text string for Legend title. Typically, the units of the data are
            used here but the type of data might also be used. Default is
            an empty string.
        base_plane: A Ladybug Plane object to note the starting point from
            where the legend will be genrated. The default is the world XY plane
            at origin (0, 0, 0).

    Properties:
        * min
        * max
        * segment_count
        * colors
        * continuous_legend
        * title
        * ordinal_dictionary
        * decimal_count
        * include_larger_smaller
        * vertical
        * base_plane
        * segment_height
        * segment_width
        * text_height
        * font

        * is_segment_count_default
        * is_title_default
        * is_base_plane_default
        * is_segment_height_default
        * is_segment_width_default
        * is_text_height_default

    Usage:

    .. code-block:: python

        lp = LegendParameters(min=0, max=100, segment_count=6)
        lp.vertical = False
        lp.segment_width = 5
    """
    __slots__ = ('_min', '_max', '_segment_count', '_colors',
                 '_continuous_legend', '_title', '_ordinal_dictionary',
                 '_decimal_count', '_include_larger_smaller', '_vertical',
                 '_base_plane', '_segment_height', '_segment_width', '_text_height',
                 '_font', '_is_segment_count_default', '_is_title_default',
                 '_is_base_plane_default', '_is_segment_height_default',
                 '_is_segment_width_default', '_is_text_height_default')

    def __init__(self, min=None, max=None, segment_count=None,
                 colors=None, title=None, base_plane=None):
        """Initialize Ladybug Legend Parameters.
        """
        self._min = None
        self._max = None
        self.min = min
        self.max = max
        self.segment_count = segment_count
        self.colors = colors
        self.title = title
        self.base_plane = base_plane

        self.continuous_legend = None
        self.ordinal_dictionary = None
        self.decimal_count = None
        self.include_larger_smaller = None
        self.vertical = None
        self.segment_height = None
        self.segment_width = None
        self.text_height = None
        self.font = None

    @classmethod
    def from_dict(cls, data):
        """Create LegendParameters from a dictionary.

    Args:
        data: A python dictionary in the following format

    .. code-block:: python

            {
            "min": -3,
            "max": 3,
            "segment_count": 7
            }
        """
        data = data.copy()  # copy to avoid mutating the input dictionary
        assert data['type'] == 'LegendParameters', \
            'Expected LegendParameters. Got {}.'.format(data['type'])
        optional_keys = ('min', 'max', 'segment_count',
                         'colors', 'continuous_legend',
                         'title', 'ordinal_dictionary',
                         'decimal_count', 'include_larger_smaller',
                         'vertical', 'base_plane', 'segment_height',
                         'segment_width', 'text_height', 'font')
        default_keys = ('is_segment_count_default', 'is_title_default',
                        'is_base_plane_default', 'is_segment_height_default',
                        'is_segment_width_default', 'is_text_height_default')
        for key in optional_keys:
            if key not in data:
                data[key] = None
        for key in default_keys:
            if key not in data:
                data[key] = False

        colors = None
        if data['colors'] is not None:
            colors = [Color.from_dict(col) for col in data['colors']]
        base_plane = None
        if data['base_plane'] is not None:
            base_plane = Plane.from_dict(data['base_plane'])

        leg_par = cls(data['min'], data['max'], data['segment_count'],
                      colors, data['title'], base_plane)
        leg_par.continuous_legend = data['continuous_legend']
        leg_par.ordinal_dictionary = data['ordinal_dictionary']
        leg_par.decimal_count = data['decimal_count']
        leg_par.include_larger_smaller = data['include_larger_smaller']
        leg_par.vertical = data['vertical']
        leg_par.segment_height = data['segment_height']
        leg_par.segment_width = data['segment_width']
        leg_par.text_height = data['text_height']
        leg_par.font = data['font']
        leg_par._is_segment_count_default = data['is_segment_count_default']
        leg_par._is_title_default = data['is_title_default']
        leg_par._is_base_plane_default = data['is_base_plane_default']
        leg_par._is_segment_height_default = data['is_segment_height_default']
        leg_par._is_segment_width_default = data['is_segment_width_default']
        leg_par._is_text_height_default = data['is_text_height_default']
        return leg_par

    @property
    def min(self):
        """Get or set legend minimum."""
        return self._min

    @min.setter
    def min(self, minimum):
        if minimum is not None:
            assert isinstance(minimum, (float, int)), \
                'Expected number for min. Got {}.'.format(type(minimum))
            if self._max is not None:
                assert minimum <= self._max, \
                    'Input min is greater than input max. {} > {}.'.format(
                        minimum, self._max)
        self._min = minimum

    @property
    def max(self):
        """Get or set legend maximum."""
        return self._max

    @max.setter
    def max(self, maximum):
        if maximum is not None:
            assert isinstance(maximum, (float, int)), \
                'Expected number for max. Got {}.'.format(type(maximum))
            if self._min is not None:
                assert maximum >= self._min, \
                    'Input max is less than input min. {} < {}.'.format(
                        maximum, self._min)
        self._max = maximum

    @property
    def segment_count(self):
        """Get or set the number of segments in the legend."""
        return self._segment_count

    @segment_count.setter
    def segment_count(self, nos):
        if nos is not None:
            assert isinstance(nos, int), \
                'Expected integer for segment_count. Got {}.'.format(type(nos))
            assert nos >= 1, 'segment_count must be greater or equal to 1.' \
                ' Got {}.'.format(nos)
            self._segment_count = nos
            self._is_segment_count_default = False
        else:
            self._segment_count = 11
            self._is_segment_count_default = True

    @property
    def colors(self):
        """Get or set the colors defining the legend."""
        return self._colors

    @colors.setter
    def colors(self, cols):
        if cols is not None:
            assert isinstance(cols, Iterable) \
                and not isinstance(cols, (str, dict, bytes, bytearray)), \
                'Colors should be a list or tuple. Got {}'.format(type(cols))
            self._colors = self._convert_colors(cols)
            assert len(self._colors) > 1, \
                'There must be at least two colors to make a legend.'
        else:
            self._colors = Colorset.original()

    @property
    def continuous_legend(self):
        """Boolean noting whether legend is drawn as a gradient or discrete segments.

        If True, the legend mesh will be drawn vertex-by-vertex resulting in a
        continuous gradient instead of discrete segments. If False, the mesh will be
        generated with one face for each of the segment_counts.
        Default: False for depicting discrete categories.
        """
        return self._continuous_legend

    @continuous_legend.setter
    def continuous_legend(self, cont_leg):
        if cont_leg is not None:
            assert isinstance(cont_leg, bool), \
                'Expected boolean for continuous_legend. Got {}.'.format(type(cont_leg))
            self._continuous_legend = cont_leg
        else:
            self._continuous_legend = False

    @property
    def title(self):
        """Get or set the text for the title of the legend."""
        return self._title

    @title.setter
    def title(self, title):
        if title is not None:
            assert isinstance(title, str), \
                'Expected string for title. Got {}.'.format(type(title))
            self._title = title
            self._is_title_default = False
        else:
            self._title = ''
            self._is_title_default = True

    @property
    def ordinal_dictionary(self):
        """Get or set an optional dictionary that maps values to text categories.

        If None, numerical values will be usedfor the legend segments. If not, text
        categories will be used and the legend will be ordinal. Note that, if the
        number if items in the dictionary are less than the segment_count, some segments
        won't receive any label. Examples for possible dictionaries include:
        {-1: 'Cold', 0: 'Neutral', 1: 'Hot'}
        {0: 'False', 1: 'True'}
        """
        return self._ordinal_dictionary

    @ordinal_dictionary.setter
    def ordinal_dictionary(self, o_dict):
        if o_dict is not None:
            assert isinstance(o_dict, dict), \
                'Expected dictionary for ordinal_dictionary. Got {}.'.format(
                    type(o_dict))
            for key in o_dict.keys():
                assert isinstance(key, int), \
                    'Expected integer for ordinal_dictionary key. Got {}.'.format(
                        type(key))
        self._ordinal_dictionary = o_dict

    @property
    def decimal_count(self):
        """Get or set an integer for the number of decimal places in the legend text.

        Default is 2. Note that this input has no bearing on the resulting legend
        text when an ordinal_dictionary is present.
        """
        return self._decimal_count

    @decimal_count.setter
    def decimal_count(self, n_dec):
        if n_dec is not None:
            assert isinstance(n_dec, int), \
                'Expected integer for decimal_count. Got {}.'.format(type(n_dec))
            self._decimal_count = n_dec
        else:
            self._decimal_count = 2

    @property
    def include_larger_smaller(self):
        """Boolean noting whether > and < should be included in legend segment text."""
        return self._include_larger_smaller

    @include_larger_smaller.setter
    def include_larger_smaller(self, lgsm):
        self._include_larger_smaller = bool(lgsm)

    @property
    def vertical(self):
        """Boolean noting whether legend is vertical (True) of horizontal (False).

        Default: True for a vertically-oriented legend.
        """
        return self._vertical

    @vertical.setter
    def vertical(self, vertical):
        if vertical is not None:
            assert isinstance(vertical, bool), \
                'Expected boolean for vertical. Got {}.'.format(
                    type(vertical))
            self._vertical = vertical
        else:
            self._vertical = True

    @property
    def base_plane(self):
        """Get or set a Ladybug Point3D for the base point of the legend."""
        return self._base_plane

    @base_plane.setter
    def base_plane(self, base_pl):
        if base_pl is not None:
            assert isinstance(base_pl, Plane), \
                'Expected Ladybug Plane for base_plane. Got {}.'.format(type(base_pl))
            self._base_plane = base_pl
            self._is_base_plane_default = False
        else:
            self._base_plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 0))
            self._is_base_plane_default = True

    @property
    def segment_height(self):
        """Get or set the height for each of the legend segments.

         Default is 1.
        """
        return self._segment_height

    @segment_height.setter
    def segment_height(self, seg_h):
        if seg_h is not None:
            assert isinstance(seg_h, (float, int)), \
                'Expected number for segment_height. Got {}.'.format(type(seg_h))
            assert seg_h > 0, 'segment_height must be greater than 0.' \
                ' Got {}.'.format(seg_h)
            self._segment_height = seg_h
            self._is_segment_height_default = False
        else:
            self._segment_height = 1
            self._is_segment_height_default = True

    @property
    def segment_width(self):
        """Get or set the width for each of the legend segments.

         Default is 1 when legend is vertical. When horizontal, the default is
         text_height * (max_number_of_digits + 2) where max_number_of_digits is
         the number of digits displaying in the legend parameter max.
        """
        if not self.vertical and self.is_segment_width_default:
            return self.text_height * 5
        return self._segment_width

    @segment_width.setter
    def segment_width(self, seg_w):
        if seg_w is not None:
            assert isinstance(seg_w, (float, int)), \
                'Expected number for segment_width. Got {}.'.format(type(seg_w))
            assert seg_w > 0, 'segment_width must be greater than 0.' \
                ' Got {}.'.format(seg_w)
            self._segment_width = seg_w
            self._is_segment_width_default = False
        else:
            self._segment_width = 1
            self._is_segment_width_default = True

    @property
    def text_height(self):
        """Get or set the height for the legend text.

        Default is 1/3 of the segment_height.
        """
        if self.is_text_height_default:
            return self.segment_height * 0.33
        return self._text_height

    @text_height.setter
    def text_height(self, txt_h):
        if txt_h is not None:
            assert isinstance(txt_h, (float, int)), \
                'Expected number for text_height. Got {}.'.format(type(txt_h))
            assert txt_h > 0, 'text_height must be greater than 0.' \
                ' Got {}.'.format(txt_h)
            self._is_text_height_default = False
        else:
            self._is_text_height_default = True
        self._text_height = txt_h

    @property
    def font(self):
        """Get or set the font for the legend text.

        Examples include "Arial", "Times New Roman", "Courier". Note that this
        parameter may not have an effect on certain interfaces that have limited
        access to fonts. Default is "Arial".
        """
        return self._font

    @font.setter
    def font(self, font):
        if font is not None:
            assert isinstance(font, str), \
                'Expected string for font. Got {}.'.format(type(font))
            self._font = font
        else:
            self._font = 'Arial'

    @property
    def is_segment_count_default(self):
        """Boolean noting whether the number of segments is defaulted."""
        return self._is_segment_count_default

    @property
    def is_title_default(self):
        """Boolean noting whether the title is defaulted."""
        return self._is_title_default

    @property
    def is_base_plane_default(self):
        """Boolean noting whether the base point is defaulted."""
        return self._is_base_plane_default

    @property
    def is_segment_height_default(self):
        """Boolean noting whether the segment height is defaulted."""
        return self._is_segment_height_default

    @property
    def is_segment_width_default(self):
        """Boolean noting whether the segment width is defaulted."""
        return self._is_segment_width_default

    @property
    def is_text_height_default(self):
        """Boolean noting whether the text height is defaulted."""
        return self._is_text_height_default

    def duplicate(self):
        """Return a copy of the current legend parameters."""
        return self.__copy__()

    def to_dict(self):
        """Get legend parameters as a dictionary."""
        base = self._base_dict()
        base['min'] = self.min
        base['max'] = self.max
        base['segment_count'] = None if self.is_segment_count_default else \
            self.segment_count
        base['ordinal_dictionary'] = self.ordinal_dictionary
        base['is_segment_count_default'] = self.is_segment_count_default
        base['type'] = 'LegendParameters'
        return base
    
    def _base_dict(self):
        """Get a dictionary with the base properties shared by all LegendParemeters."""
        title = None if self.is_title_default else self.title
        base_plane = None if self.is_base_plane_default else self.base_plane.to_dict()
        seg_h = None if self.is_segment_height_default else self.segment_height
        seg_w = None if self.is_segment_width_default else self.segment_width
        txt_h = None if self.is_text_height_default else self.text_height
        return {
            'colors': [col.to_dict() for col in self.colors],
            'continuous_legend': self.continuous_legend, 'title': title,
            'decimal_count': self.decimal_count,
            'include_larger_smaller': self.include_larger_smaller,
            'vertical': self.vertical,
            'base_plane': base_plane,
            'segment_height': seg_h, 'segment_width': seg_w,
            'text_height': txt_h, 'font': self.font,
            'is_title_default': self.is_title_default,
            'is_base_plane_default': self.is_base_plane_default,
            'is_segment_height_default': self.is_segment_height_default,
            'is_segment_width_default': self.is_segment_width_default,
            'is_text_height_default': self.is_text_height_default
        }

    @staticmethod
    def _convert_colors(cols):
        """Convert a list of colors into ladybug Color objects."""
        try:
            cols = tuple(col if isinstance(col, Color) else Color(
                col.R, col.G, col.B) for col in cols)
        except Exception:
            try:
                cols = tuple(Color(col.Red, col.Green, col.Blue)
                                for col in cols)
            except Exception:
                raise ValueError("{} is not a valid list of colors".format(cols))
        return cols

    def __copy__(self):
        new_par = LegendParameters(self.min, self.max, self.segment_count,
                                   self.colors, self.title, self.base_plane)
        new_par._continuous_legend = self._continuous_legend
        new_par._ordinal_dictionary = self._ordinal_dictionary
        new_par._decimal_count = self._decimal_count
        new_par._include_larger_smaller = self._include_larger_smaller
        new_par._vertical = self._vertical
        new_par._segment_height = self._segment_height
        new_par._segment_width = self._segment_width
        new_par._text_height = self._text_height
        new_par._font = self._font
        new_par._is_segment_count_default = self._is_segment_count_default
        new_par._is_title_default = self._is_title_default
        new_par._is_base_plane_default = self._is_base_plane_default
        new_par._is_segment_height_default = self._is_segment_height_default
        new_par._is_segment_width_default = self._is_segment_width_default
        new_par._is_text_height_default = self._is_text_height_default
        return new_par

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        """Legend parameter representation."""
        min = self.min if self.min is not None else '[default]'
        max = self.max if self.max is not None else '[default]'
        seg = '[default]' if self.is_segment_count_default \
            else self.segment_count
        title = '[default]' if self.is_title_default else self.title
        base_pt = '[default]' if self.is_base_plane_default else self.base_plane.o
        seg_h = '[default]' if self.is_segment_height_default else self.segment_height
        seg_w = '[default]' if self.is_segment_width_default else self.segment_width
        txt_h = '[default]' if self.is_text_height_default else self.text_height
        return 'Legend Parameters\n minimum: {}\n maximum: {}\n segments: {}\n' \
            ' colors:\n  {}\n continuous legend: {}\n' \
            ' title: {}\n ordinal text: {}\n number decimals: {}\n' \
            ' include < >: {}\n vertical: {}\n base point:\n  {}\n' \
            ' segment height: {}\n segment width: {}\n' \
            ' text height: {}\n font: {}'.format(
                min, max, seg, '\n  '.join([str(c) for c in self.colors]),
                self.continuous_legend, title,
                self.ordinal_dictionary, self.decimal_count,
                self.include_larger_smaller, self.vertical,
                base_pt, seg_h, seg_w, txt_h, self.font)


class LegendParametersCategorized(LegendParameters):
    """Ladybug legend parameters used to customize legends.

    These legend parameters have more limitations than the base LegendParameters
    class. However, these legend parameters will do auto-categorization of data,
    binning values into groups based on custom ranges.

    Args:
        domain: A list of one or more numbers noting the bondaries of the data
            categories. For example, [100, 2000] creates three categories of
            (<100, 100-2000, >2000). Values must always be ordered from lowest
            to highest.
        colors: An list of color objects with a length equal to the number of items
            in the domain + 1. These are used to color each of the categories of data.
        category_names: An optional list of text strings with a length equal to the
            colors. These will be used to name each of the categories in the legend.
            If None, the legend text will simply mark the numerical ranges of the
            categories. (Default: None).
        title: Text string for Legend title. Typically, the units of the data are
            used here but the type of data might also be used. Default is
            an empty string.
        base_plane: A Ladybug Plane object to note the starting point from
            where the legend will be genrated. The default is the world XY plane
            at origin (0, 0, 0).

    Properties:
        * domain
        * colors
        * category_names
        * continuous_colors
        * continuous_legend
        * title
        * ordinal_dictionary
        * decimal_count
        * include_larger_smaller
        * vertical
        * base_plane
        * segment_height
        * segment_width
        * text_height
        * font
        * min
        * max
        * segment_count

        * is_title_default
        * is_base_plane_default
        * is_segment_height_default
        * is_segment_width_default
        * is_text_height_default
    """
    __slots__ = ('_domain', '_category_names', '_continuous_colors',)

    def __init__(self, domain, colors, category_names=None, title=None, base_plane=None):
        """Initialize Ladybug Legend Parameters Categorized."""
        # set the domain after verifying that it is correct
        assert isinstance(domain, Iterable) \
            and not isinstance(domain, (str, dict, bytes, bytearray)), \
            'Domain should be a list or tuple. Got {}'.format(type(domain))
        self._domain = tuple(float(x) for x in sorted(domain))
        assert len(self._domain) > 0, \
            'LegendParametersCategorized domain must have at least one value.'
        self._min = self._domain[0]
        self._max = self._domain[-1]
        self._segment_count = len(self._domain) + 1

        # set the other input arguments
        self.colors = colors
        self.category_names = category_names
        self.title = title
        self.base_plane = base_plane

        ## set all of the other inputs to None for now.
        self.continuous_colors = None
        self.continuous_legend = None
        self.decimal_count = None
        self.include_larger_smaller = None
        self.vertical = None
        self.segment_height = None
        self.segment_width = None
        self.text_height = None
        self.font = None

        # properties that have no meaning for this class
        self._ordinal_dictionary = None
        self._is_segment_count_default = True

    @classmethod
    def from_dict(cls, data):
        """Create LegendParametersCategorized from a dictionary.

    Args:
        data: A python dictionary in the following format

    .. code-block:: python

            {
            "domain": [100, 2000],
            "colors": [{'r': 0, 'g': 0, 'b': 0},
                       {'r': 0, 'g': 0, 'b': 100},
                       {'r': 255, 'g': 0, 'b': 0}],
            "category_names": ["low", "desired", "too much"]
            }
        """
        data = data.copy()  # copy to avoid mutating the input dictionary
        assert data['type'] == 'LegendParametersCategorized', \
            'Expected LegendParametersCategorized. Got {}.'.format(data['type'])
        optional_keys = ('category_names', 'continuous_legend', 'continuous_colors',
                         'title', 'ordinal_dictionary',
                         'decimal_count', 'include_larger_smaller',
                         'vertical', 'base_plane', 'segment_height',
                         'segment_width', 'text_height', 'font')
        default_keys = ('is_title_default', 'is_base_plane_default',
                        'is_segment_height_default', 'is_segment_width_default',
                        'is_text_height_default')
        for key in optional_keys:
            if key not in data:
                data[key] = None
        for key in default_keys:
            if key not in data:
                data[key] = False

        colors = [Color.from_dict(col) for col in data['colors']]
        base_plane = None
        if data['base_plane'] is not None:
            base_plane = Plane.from_dict(data['base_plane'])

        leg_par = cls(data['domain'], colors, data['category_names'],
                      data['title'], base_plane)
        leg_par.continuous_colors = data['continuous_colors']
        leg_par.continuous_legend = data['continuous_legend']
        leg_par.decimal_count = data['decimal_count']
        leg_par.include_larger_smaller = data['include_larger_smaller']
        leg_par.vertical = data['vertical']
        leg_par.segment_height = data['segment_height']
        leg_par.segment_width = data['segment_width']
        leg_par.text_height = data['text_height']
        leg_par.font = data['font']
        leg_par._is_title_default = data['is_title_default']
        leg_par._is_base_plane_default = data['is_base_plane_default']
        leg_par._is_segment_height_default = data['is_segment_height_default']
        leg_par._is_segment_width_default = data['is_segment_width_default']
        leg_par._is_text_height_default = data['is_text_height_default']
        return leg_par

    @property
    def domain(self):
        """Get or set a list of numbers noting the boundaries of the data categories."""
        return self._domain

    @domain.setter
    def domain(self, dom):
        assert isinstance(dom, Iterable) \
            and not isinstance(dom, (str, dict, bytes, bytearray)), \
            'Domain should be a list or tuple. Got {}'.format(type(dom))
        self._domain = tuple(float(x) for x in sorted(dom))
        assert len(self._domain) == len(self._colors) - 1, 'The length of domain must' \
            'be one less than length of the colors for a LegendParametersCategorized.' \
            '{} != {} - 1'.format(len(self._domain), len(self._colors))
        self._min = self._domain[0]
        self._max = self._domain[-1]
        self._segment_count = len(self._domain) + 1

    @property
    def colors(self):
        """Get or set the colors defining the legend."""
        return self._colors

    @colors.setter
    def colors(self, cols):
        assert isinstance(cols, Iterable) \
            and not isinstance(cols, (str, dict, bytes, bytearray)), \
            'Colors should be a list or tuple. Got {}'.format(type(cols))
        self._colors = self._convert_colors(cols)
        assert len(self._colors) == len(self._domain) + 1, 'The length of colors must ' \
            'be one more than the length of domain for a LegendParametersCategorized.' \
            '{} != {} + 1'.format(len(self._colors), len(self._domain))

    @property
    def category_names(self):
        """Get or set a list of text for the names of the categories."""
        if self._category_names:  # user-specified category names
            return self._category_names
        # generate the category names based on the domain
        format_str = '%.{}f'.format(self.decimal_count)
        nums = [format_str % x for x in self.domain]
        mid_nums = tuple('{} - {}'.format(nums[i], nums[i + 1])
                         for i in xrange(len(nums) - 1))
        if self.include_larger_smaller:
            return ('<' + nums[0],) + mid_nums + ('>' + nums[-1],)
        return (nums[0],) + mid_nums + (nums[-1],)

    @category_names.setter
    def category_names(self, categories):
        if categories is not None:
            assert isinstance(categories, Iterable) \
                and not isinstance(categories, (str, dict, bytes, bytearray)), \
                'Category names should be a list or tuple. Got {}'.format(type(categories))
            self._category_names = tuple(str(x) for x in categories)
            assert len(self._category_names) == len(self._domain) + 1, 'The length of ' \
                'category_names must be one more than length of the colors for a ' \
                'LegendParametersCategorized.{} != {} - 1'.format(
                    len(self._domain), len(self._colors))
        else:
            self._category_names = None

    @property
    def continuous_colors(self):
        """Boolean noting whether colors generated are continuous or discrete.

        If True, the colors generated from the corresponding legend will be in a
        continuous gradient. If False, they will be categorized in incremental
        groups according to the segment_count. (Default: False).
        """
        return self._continuous_colors

    @continuous_colors.setter
    def continuous_colors(self, cont_cols):
        if cont_cols is not None:
            assert isinstance(cont_cols, bool), \
                'Expected boolean for continuous_colors. Got {}.'.format(type(cont_cols))
            self._continuous_colors = cont_cols
        else:
            self._continuous_colors = False

    @property
    def include_larger_smaller(self):
        """Boolean noting whether > and < should be included in legend segment text."""
        return self._include_larger_smaller

    @include_larger_smaller.setter
    def include_larger_smaller(self, lg_sm):
        if lg_sm is not None:
            assert isinstance(lg_sm, bool), 'Expected boolean for ' \
                'include_larger_smaller. Got {}.'.format(type(lg_sm))
            self._include_larger_smaller = lg_sm
        else:
            self._include_larger_smaller = True

    @property
    def min(self):
        """Get legend minimum. This is derived from the domain."""
        return self._min

    @property
    def max(self):
        """Get legend maximum. This is derived from the domain."""
        return self._max

    @property
    def segment_count(self):
        """Get the number of segments in the legend.
        
        This is always equal to one more than the length of the domain."""
        return self._segment_count

    @property
    def ordinal_dictionary(self):
        """Always None for a LegendParametersCategorized."""
        return self._ordinal_dictionary

    def to_dict(self):
        """Get legend parameters categorized as a dictionary."""
        base = self._base_dict()
        base['domain'] = self.domain
        base['category_names'] = self.category_names
        base['continuous_colors'] = self.continuous_colors
        base['type'] = 'LegendParametersCategorized'
        return base

    def __copy__(self):
        new_par = LegendParametersCategorized(
            self._domain, self._colors, self._category_names, self.title,
            self.base_plane)
        new_par._continuous_colors = self._continuous_colors
        new_par._continuous_legend = self._continuous_legend
        new_par._decimal_count = self._decimal_count
        new_par._include_larger_smaller = self._include_larger_smaller
        new_par._vertical = self._vertical
        new_par._segment_height = self._segment_height
        new_par._segment_width = self._segment_width
        new_par._text_height = self._text_height
        new_par._font = self._font
        new_par._is_title_default = self._is_title_default
        new_par._is_base_plane_default = self._is_base_plane_default
        new_par._is_segment_height_default = self._is_segment_height_default
        new_par._is_segment_width_default = self._is_segment_width_default
        new_par._is_text_height_default = self._is_text_height_default
        return new_par

    def __repr__(self):
        """Legend parameter representation."""
        title = '[default]' if self.is_title_default else self.title
        base_pt = '[default]' if self.is_base_plane_default else self.base_plane.o
        seg_h = '[default]' if self.is_segment_height_default else self.segment_height
        seg_w = '[default]' if self.is_segment_width_default else self.segment_width
        txt_h = '[default]' if self.is_text_height_default else self.text_height
        return 'Legend Parameters Categorized\n domain: {}\n colors: {}\n' \
            ' category names\n  {}\n continuous legend: {}\n' \
            ' title: {}\n ordinal text: {}\n number decimals: {}\n' \
            ' include < >: {}\n vertical: {}\n base point:\n  {}\n' \
            ' segment height: {}\n segment width: {}\n' \
            ' text height: {}\n font: {}'.format(
                self.domain, 
                '\n  '.join([str(c) for c in self.colors]),
                '\n  '.join([str(c) for c in self.category_names]),
                self.continuous_legend, title,
                self.ordinal_dictionary, self.decimal_count,
                self.include_larger_smaller, self.vertical,
                base_pt, seg_h, seg_w, txt_h, self.font)



# coding=utf-8
"""Energy data type."""
from __future__ import division

from .base import DataTypeBase
from .energyintensity import EnergyIntensity


class Energy(DataTypeBase):
    """Energy
    """
    _units = ('kWh', 'kBtu', 'Wh', 'Btu', 'MMBtu', 'J', 'kJ', 'MJ', 'GJ',
              'therm', 'cal', 'kcal')
    _si_units = ('kWh', 'Wh', 'J', 'kJ', 'MJ', 'GJ')
    _ip_units = ('kBtu', 'Btu', 'MMBtu', 'therm')
    _abbreviation = 'E'
    _point_in_time = False
    _cumulative = True
    _normalized_type = EnergyIntensity

    def _kWh_to_kBtu(self, value):
        return value * 3.41214

    def _kWh_to_Wh(self, value):
        return value * 1000.

    def _kWh_to_Btu(self, value):
        return value * 3412.14

    def _kWh_to_MMBtu(self, value):
        return value * 0.00341214

    def _kWh_to_J(self, value):
        return value * 3600000.

    def _kWh_to_kJ(self, value):
        return value * 3600.

    def _kWh_to_MJ(self, value):
        return value * 3.6

    def _kWh_to_GJ(self, value):
        return value * 0.0036

    def _kWh_to_therm(self, value):
        return value * 0.0341214

    def _kWh_to_cal(self, value):
        return value * 860421.

    def _kWh_to_kcal(self, value):
        return value * 860.421

    def _kBtu_to_kWh(self, value):
        return value / 3.41214

    def _Wh_to_kWh(self, value):
        return value / 1000.

    def _Btu_to_kWh(self, value):
        return value / 3412.14

    def _MMBtu_to_kWh(self, value):
        return value / 0.00341214

    def _J_to_kWh(self, value):
        return value / 3600000.

    def _kJ_to_kWh(self, value):
        return value / 3600.

    def _MJ_to_kWh(self, value):
        return value / 3.6

    def _GJ_to_kWh(self, value):
        return value / 0.0036

    def _therm_to_kWh(self, value):
        return value / 0.0341214

    def _cal_to_kWh(self, value):
        return value / 860421.

    def _kcal_to_kWh(self, value):
        return value / 860.421

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('kWh', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'Wh':
            return self.to_unit(values, 'Btu', from_unit), 'Btu'
        else:
            return self.to_unit(values, 'kBtu', from_unit), 'kBtu'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'Btu':
            return self.to_unit(values, 'Wh', from_unit), 'Wh'
        else:
            return self.to_unit(values, 'kWh', from_unit), 'kWh'
